#include "Input.h"
#include "LoopList.h"


void ReadInput::AddPair(int m, int n, char c)
{

  //m-current position, n-its pair according to sequence array, c-its corresponding base type
  int size = 0;
  c = toupper(c);
  CSequence[m] = c;
  switch (c){
  case 'A':
    type[m] = A;
    break;
  case 'C':
    type[m] = C;
    break;
  case 'G':
    type[m] = G;
    break;
    //case 'T':
    // type[m] = T;
    // break;
  case 'U': 
    type[m] = U;
    break;
  };

  if (size < m)
    size = m;
  if (size < n)
    size = n;
  if ((m > 0) && (n > 0)){
    Sequence[m] = n;
    Sequence[n] = m;
  };

  if (m == 0)
    Sequence[n] = -1;
  if (n == 0)
    Sequence[m] = -1;

  if (Size < size)
    Size = size;
};

ReadInput::ReadInput(char * fseq, char * fbpseq)
{
  // initializing given two files where one contains the primary sequence and the other
  // base pairs in the sequence

  for (int i = 0; i < MaxN; i++)
    {
      Sequence[i] = -1;
      CSequence[i] = ' ';
      Next[i] = 0;
      Prev[i] = 0;
      type[i] = 0;
      loops[i].pair = 0;
      looplists[i] = NULL;
    };

  CSequence[0] = ' ';

  FILE * fileSeq = fopen(fseq, "r+");
  FILE * fileBpseq = fopen(fbpseq, "r+");
  int  m, n;

  char c[MaxN+1];
  int size = 0;
  while(!feof(fileSeq)) 
    {
      fscanf(fileSeq, "%s", c);
    }

  int i=0;
  while (!feof(fileBpseq))
    {
      fscanf(fileBpseq, "%d %d\n", &m,  &n);
      AddPair(m, n, c[i]);
      i++;
    };

  fclose(fileSeq);
  fclose(fileBpseq);

  int Last = 0;
  for (int m = 1; m <= Size; m++)
    if (Sequence[m] > 0)
      {
	loops[m].pair = BasePair(m);
	Prev[m] = Last;
	Next[Last] = m;
	Last = m;
      };
};

ReadInput::ReadInput(int n, vector <int> sequence, vector <char> csequence, vector <int> to_remove){
  // Initializing given the base pairs in the sequence (sequence) and primary sequence (csequence)

  //memset(tot_bases, 0, sizeof(tot_bases));
  //memset(tot_pairs, 0, sizeof(tot_pairs));
  
  
  where_at=0;
  not_finished=true;
  to_be_removed=to_remove;

  tot_Bases=tot_Pairs=tot_pBases=tot_pPairs=0;

  basesChar[0] = 'A';
  basesChar[1] = 'U';
  basesChar[2] = 'C';
  basesChar[3] = 'G';

  pairsChar[0] = "AA";
  pairsChar[1] = "UU";
  pairsChar[2] = "CC";
  pairsChar[3] = "GG";
  pairsChar[4] = "AC";
  pairsChar[5] = "AG";
  pairsChar[6] = "UC";
  pairsChar[7] = "UG";
  pairsChar[8] = "AU";
  pairsChar[9] = "CG";
  pairsChar[10] = "other base pair";

  tot_bands=tot_pseudo=tot_stems=tot_hairpins=tot_bulges=tot_inter=tot_multi=tot_external=tot_stacks=tot_domains=0;

  tot_pstems=tot_phairpins=tot_pbulges=tot_pinter=tot_pmulti=tot_pstacks=0;
    
/*  
  EC1[0]="multi ";
  EC1[1]="hairpin ";
  EC1[2]="internal ";
  EC1[3]="bulge ";
  EC1[4]="stack ";
  EC1[5]="free ";
  EC1[6]="pseudo multi ";
  EC1[7]="pseudo hairpin ";
  EC1[8]="pseudo internal ";
  EC1[9]="pseudo bulge ";
  EC1[10]="pseudo stack ";
  
  EC2[0]="/external begin";
  EC2[1]="/external end";
  EC2[2]="begin";
  EC2[3]="end";
  EC2[4]="base";
  
  //now initiate everything to be a free base 
  for(int i=0; i<n; i++)
  {
  Enriched[i][0]=5;
  Enriched[i][1]=4;
  }
  */
  
 
 avg_stem=avg_band=avg_h=avg_i=avg_b=avg_m=0;
 max_stem=max_band=min_stem=min_band=0;
 max_h=min_h=max_i=min_i=0;
 max_b=min_b=max_m=min_m=0;
 num_length_i=num_length_b=num_length_m=0;  
 
 
 tot_fbh=tot_fbi=tot_fbb=tot_fbm=tot_fbe=0;
  
 for(int i=0; i<MaxBases; i++)
  fbh[i]=fbi[i]=fbb[i]=fbm[i]=fbe[i]=fbpl[i]=fbwb[i]=tot_bases[i]=tot_pbases[i]= 0;
 
 for(int i=0; i<MaxPairs; i++)
   cbph[i]=cbpi[i]=cbpb[i]=cbpm[i]=cbpe[i]=tot_pairs[i]=tot_ppairs[i]= 0;
  
 tot_cbph=tot_cbpi=tot_cbpb=tot_cbpm=tot_cbpe=tot_fbpl=tot_fbwb=0;
  
  //end of Vera's code
  
  
  
  for (int i = 0; i < MaxN; i++)
    {
      Sequence[i] = -1;
      CSequence[i] = ' ';
      loops[i].pair = 0;
      Next[i] = 0;
      Prev[i] = 0;
      looplists[i] = NULL;
      type[i] = 0;
    };

  CSequence[0] = ' ';
  for (int i = 0; i < n; i++)
    {
      //cout << "Adding pair: " << i+1 << " " << sequence[i] << endl;
      AddPair(i+1, sequence[i], csequence[i]);
    };

  Size = n;
  int Last = 0;

  for (int m = 1; m <= Size; m++)
    if (Sequence[m] > 0)
      {
	loops[m].pair = BasePair(m);
	Prev[m] = Last;
	Next[Last] = m;
	Last = m;
      };

  for (int i = 0; i <= Size; i++)
    ClosedRegions[i] = NULL;
};

ReadInput::ReadInput(int size, const unsigned char* baseSequence, const unsigned short* pairRefSequence)
{
  //printf("[ReadInput] %s\n", baseSequence);
  for (int i = 0; i < MaxN; i++)
    {
      Sequence[i] = -1;
      CSequence[i] = ' ';
      looplists[i] = NULL;
      loops[i].pair = 0;
      Next[i] = 0;
      Prev[i] = 0;
      type[i] = 0;
      //cannot_add_dangling[i] = 0;
    };

  char c;

  for (int i = 1; i <= size; i++)
    {
      switch ((int)baseSequence[i])
	{
	case A:
	  c = 'A';
	  break;
	case C:
	  c = 'C';
	  break;
	case G:
	  c = 'G';
	  break;
	case U:
	  c = 'U';
	  break;
	};
      AddPair(i, (int)pairRefSequence[i], c);
    };

  Size = size;
  int Last = 0;

  for (int m = 1; m <= Size; m++)
    if (Sequence[m] > 0)
      {
	loops[m].pair = BasePair(m);
	Prev[m] = Last;
	Next[Last] = m;
	Last = m;
      };

  for (int i = 0; i <= Size; i++)
    ClosedRegions[i] = NULL;
};


int ReadInput::BasePair(int a)
{
  return Sequence[a];
};

ReadInput::~ReadInput()
{
};



