#include "Loop.h"
//#define NULL 0

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

Loop::Loop(int b, int e,  ReadInput * R, Bands * B, Stack * S)
{
  //addint a new loop(closed region) and initializing it
  ILoops = NULL;
  MLoops = NULL;
  St = S;
  LeftSibling = NULL;
  RightChild = NULL;
  Parent = NULL;
  PseudoParent = NULL;
  begin = b;
  end = e;
  Input = R;
  bandpattern        = B;
  type = external;
  NumberOfChildren= 0;
  NumberOfUnpaird= 0;

  nested = nothing;
  NumberOfUnBandChild = 0;
  NumberOfBandChild = 0;
  
  for(int i=0; i<MaxBases; i++)
    fb[i]=0;
  
  for(int i=0; i<MaxPairs; i++)
    cbp[i]=0;
  
  tot_fb=tot_cbp=0;
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

Loop::~Loop()
{
  delete RightChild;
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::WhereLocated()
{
  // It figures out the location of the loop in a pseudoknot :"inBand", "unBand".
  // Consider that spanBand loops do not have any node in the tree (as they are not closed region)
  // and will be discoverd in FindInnerLoops (which uses LoopList)

  int previousRegion;

  while (true)
    {
      if (( end <= bandpattern->pattern[Parent ->CurrentBandRegion].OtherBorder) &&
          (end >= Parent->CurrentBandRegion) )
        {
          if (begin >= Parent->CurrentBandRegion) 
            {
              nested = inBand;
              return;
            }
          //else
          // else  crossBand
        }
      else 
        {
          previousRegion = bandpattern->pattern[Parent ->CurrentBandRegion].prev;
          if (( end > bandpattern->pattern[previousRegion].OtherBorder) &&
              (end < Parent->CurrentBandRegion) ) //sure: begin >bandpattern->pattern[previousRegion].OtherBorder
            {
              nested = unBand;
              return;
            }
          else {
            Parent ->CurrentBandRegion = previousRegion;
          }
        }
    }
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::FindInnerLoops(int border1, int border2)                //Finds all Interior nodes in the band
{
  //MultiAllLoops = new LoopList(Input, 0, MaxN);
  int i=border1;
  while (i <= border2)
    {
      int j = bandpattern->pattern[i].OtherBorder;

      if  (Input->BasePair(i) > i)
        {
          int ip = Input->BasePair(i);
          int jp = Input->BasePair(j);
          //[i,j]|_| [jp,ip] is a band and i < i'

          int x = j;
          while (x > i)
            {
              //for (int x = j; x > i; x--){
              int y = St->PrevInStack[x];
              int xp = Input->BasePair(x);
              int yp = Input->BasePair(y);
              /*
                printf("[InnterLoops] Checking [%d,%d]-[%d,%d]\n", y, x, xp, yp); fflush(stdout);
                printf("PrevIndex[%d] = %d\n", yp, St->PrevInStack[yp]);
              */

              //Check to see if x--y |_| yp--xp is a loop

              if ((xp >= jp) && (yp >= jp) && (xp >= St->PrevInStack[yp]))
                {
                  if ((x == Input->Next[y]) && (yp == Input->Next[xp]))
                    {                //InteriorLoop
                      if  (Input->BasePair(i) > i)
                        {
                          //InteriorLoops->Add(y, x);

                          T_IntList * IntList = new T_IntList;
                          IntList->Num = y;
                          IntList->Next = NULL;
                          Input->looplists[y] = new LoopList(Input, y, x);
                          IntList->Next = ILoops;
                          ILoops = IntList;
                        };
                    }

                  else
                    {                                                                                //MultiLoop;
                      //MultiLoops->Add(y, x);
                      T_IntList * IntList = new T_IntList;
                      IntList->Num = y;
                      IntList->Next = NULL;
                      Input->looplists[y] = new LoopList(Input, y, x);
                      Input->looplists[y]->FindChildren();
                      IntList->Next = MLoops;
                      MLoops = IntList;
                    };
                };
              x = y;
            };
        }
      i = bandpattern->pattern[i].next;
    }
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::PropagateNested()
{
  // When a node corresponds to a pseudoknotLoop, this function is called to
  // find out the location of its children ("inBand",...) using WhereLocated function
  // and also calculate the number of children in each of these types.
  if (type == pseudo)
    return;

  Loop * L = RightChild;
  while (L!=NULL)
    {
      L->nested = inMulti; //inmulti for now, might be nothing (it means we dont need propagate)
      L->PropagateNested();
      L= L->LeftSibling;
    }
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::PseudoNestedCheck()
{
  if (type != pseudo)
    return;

  Loop * L = RightChild;
  while (L != NULL)
    {
      L ->WhereLocated();

      if (L->nested == inBand)
        {
          NumberOfBandChild = NumberOfBandChild + 1;
        }

      else if (L->nested == unBand)
        NumberOfUnBandChild = NumberOfUnBandChild +1;
      //L ->PropagateNested();
      L = L->LeftSibling;
    };
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::loopSetType()
{
  if (Input->BasePair(begin) != end)
    {
      type = pseudo;
      Input->tot_pseudo++;
      Num=Input->tot_pseudo;
      Name = "pseudo region ";
      Input->left[Input->tot_pseudo-1]= begin;
      Input->right[Input->tot_pseudo-1]= end;      
      return;
    };
  //        hairpin,        interior,        multi,        external,        pseudo
  if (!RightChild)
    {
      type = hairpin;
      Input->tot_hairpins++;
      Num=Input->tot_hairpins;
      Name = "hairpin loop ";
      Input->tot_stacks++;
      return;
    };

  if ((!RightChild->LeftSibling))// && (Input->BasePair(RightChild->begin) == RightChild->end))
    {
      if ((RightChild->begin == begin + 1) && (RightChild->end == end - 1 ))
        {
          type = stackloop;
          Input->tot_stacks++;
          Num=Input->tot_stacks;
          Name = "stack ";
          int border1=0;
          int border2=0;         
          if(begin-1 >= 0)
            border1=begin-1;
          if(end+1 < MaxN)
            border2=end+1;

          if(Input->BasePair(border2) != border1) //ie if the one above wasn't a stacked pair that contained the one we have now
            {
              Input->tot_stems++;
              Num=Input->tot_stems;
              Name = "stem ";
            }
          return;
        }
      else if((RightChild->begin == begin+1)||(RightChild->end == end-1))
        {
          type = bulge;
          Input->tot_bulges++;
          Num=Input->tot_bulges;
          Name = "bulge ";
          Input->tot_stacks++;
          return;
        }
      else
        {
          type = interior;
          Input->tot_inter++;
          Num=Input->tot_inter;
          Name = "internal loop ";
          Input->tot_stacks++;
          return;
        };
      
    };

  if ((RightChild->LeftSibling))// || (Input->BasePair(RightChild->begin) != RightChild->end)) 
    {
      type = multi;
      Input->tot_multi++;
      Num=Input->tot_multi;
      Name = "multi loop ";
      Input->tot_stacks++;
      return;
    };


  type = external;
  Input->tot_external++;
  Num=Input->tot_external;
  Name = "external loop ";
  return;
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::setPseudoParent()
{
  // This function is not useful now! It might be usefull later for other energy calculation models.
  // but for now we do not use it.
  if (Parent)
    if (Parent->type == pseudo)
      PseudoParent = Parent;
    else
      PseudoParent = Parent->PseudoParent;

  Loop * L = RightChild;
  if (!L)
    return;

  while (L)
    {
      L->setPseudoParent();
      L = L->LeftSibling;

    };
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

//Number of Unpaired bases in range [from+1, to-1]
int UnpairedBases(ReadInput* Input, int from, int to)
{
  int k = 0;
  for (int j = from; j < to; j++)
    if (Input->BasePair(j) <= 0)
      k++;

  return k;
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::addLoop(int begin, int end)
{
  Loop *L1 = new Loop(begin, end, Input, bandpattern, St);
  Input->ClosedRegions[begin] = L1;

  //To cacluate         NumberOfUnpaird bases
  int last = end;
  int first = begin;
  L1->Parent = this;

  if (!RightChild)
    {
      RightChild = L1;
      if (begin != end)
        L1->NumberOfUnpaird = UnpairedBases(Input, first, last);
      L1->loopSetType();
      L1->LoopFindBands();
      L1 ->PseudoNestedCheck();

      return;
    };


  Loop *L = RightChild;
  int HaveChild = (L->begin >= begin);
  if (HaveChild)
    L1->RightChild = L; 
  Loop *L2 = L;
  Input->loops[begin].num_branches = 0;
  Input->loops[begin].pseudo_num_branches = 0;


  while (L && (L->begin >= begin))
    {
      Input->loops[begin].bri[Input->loops[begin].num_branches++] = L->begin;
      if (L->type == pseudo)
        {
          Input->loops[begin].pseudo_num_branches += 2;
        }
      else
        {
          Input->loops[begin].pseudo_num_branches += 1; 
        }

      if (begin != end)
        {
          L1->NumberOfUnpaird += UnpairedBases(Input, L->end, last); 
        }
      last = L->begin;

      L1->NumberOfChildren++;
      L2 = L;
      L->Parent = L1;
      L = L->LeftSibling;
    };

  for (int index = 0; index < Input->loops[begin].num_branches; index++)
    {
      int index2 = Input->loops[begin].num_branches - index - 1;
      if (index < index2)
        {
          int x = Input->loops[begin].bri[index];
          Input->loops[begin].bri[index] = Input->loops[begin].bri[index2];
          Input->loops[begin].bri[index2] = x;
        };
    };
  if (begin != end)
    L1->NumberOfUnpaird += UnpairedBases(Input, first, last);
  if (L)
    L1->LeftSibling = L;
  if (L2 && HaveChild)
    L2->LeftSibling = NULL;

  RightChild = L1;
  L1->loopSetType();
  L1->LoopFindBands();
  L1 ->PseudoNestedCheck();
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::LoopFindBands()
{
  if ((type != pseudo) || (begin == 0))
    {
      if (begin > 0)
        bandpattern->Update_links(bandpattern->Prev(begin) , bandpattern->Next(end));
      return;
    };
  //JUST GET THE BAND BORDERS HERE.
  bandpattern->aux_Find_bands(begin, end, &NumberOfBands,&CurrentBandRegion, bands_remove, bandborders);
  bandpattern->Update_links(bandpattern->Prev(begin) , bandpattern->Next(end));
  
  FindInnerLoops(begin, end);

}




//AFTER THIS: ALL THE FUNCTIONS WERE WRITTEN BY VERA (PRINT WAS MODIFIED BEYOND RECONGITION, A COUPLE OF FUNCTIONS WERE INHERITED FROM MIRELA, THEY ARE MARKED IN .H FILE)


//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

// This function has been much modified by Vera 

void Loop::Print(int i)
{

  /*
    for (int j = 0; j < i; j++)
    printf(".");
  */  
    
  if (Parent)
    {
      if(type==stackloop) 
        {
          // need: length (in basepairs), positions for two closing base pairs, connections, %s and absolute
          // numbers for the make up (just bases and base pairs), contex for non cannonical base pairs, 
          // plus free energy, and connection to two things (parent and child should do the trick)
          
          // Mirela APril 14, 2005: The second condition shouldn't be there - its doesn't print stems 
          //    which follow immediately a bulge with no free bases on the left
          if((Parent->type != stackloop) || (i==0)) 
          //if(((Parent->type != stackloop)&&(Parent->begin != begin-1)) || (i==0)) //need to check that its parent is not a stack that has already been reported...
            {
              printf("closed region [%d, %d]\n\n",begin, end);
              cout << "***" << Name << Num << "***\n" ;
              Loop * STK = StemConnections(i);
              cout << "Free Stem Energy: " << StemFreeEnergy(begin, STK->begin) << endl;
              Numbers(begin,STK->begin,true);
              cout << endl << "*******************************************************************************\n\n";
            }         
        }
      else if(type==hairpin)
        {
          // need: length (in unpaired base pairs), position for closing pair, types according to motif and # of
          // free bases, plus connection to one thing, %s and absolute numbers for the make up
          printf("closed region [%d, %d]\n\n",begin, end);
          cout << "***" << Name << Num << "***\n";
          HairpinConnections(i);

          HairpinType(); //this is according to a motif
          if(NumberOfUnpaird==3)
            cout << "Note: this is a tri hairpin\n";
          if(NumberOfUnpaird==4)
            cout << "Note: this is a tetra hairpin\n";
          cout << endl;
          
          //want to take into account stacked pairs of length 1, thus... this might be a stacked pair, too
          if(Parent->type != stackloop)
            SingleStacked(i);
            
          cout << "******************************************************************************\n\n";
        }
      else if(type==bulge) 
        {
          //need: length (in unpaired bases), position of 2 closing pairs, connection to two stems,
          // plus %s and absolute numbers for the make up
          printf("closed region [%d, %d]\n\n",begin, end);
          cout << "***" << Name << Num << "***\n";
          MultiConnections(false, i);
          cout << endl;
    
          //want to take into account stacked pairs of length 1, thus... this might be a stacked pair, too
          if(Parent->type != stackloop)
            SingleStacked(i);
            
          cout << "******************************************************************************\n\n";
        }
    
      else if(type==interior)
        {
          //need: length (in twos, of unpaired bases), everything else same as with bulge, plus
          // asymmentry measure (difference and ratio being reported)
          printf("closed region [%d, %d]\n\n",begin, end);
          cout << "***" << Name << Num << "***\n";
          MultiConnections(false, i);
          cout << endl;
          
          //want to take into account stacked pairs of length 1, thus... this might be a stacked pair, too
          if(Parent->type != stackloop)
            SingleStacked(i);  
            
          cout << "******************************************************************************\n\n";  
        } 
      else if(type==multi)
        {
          //need: lengths of unpaired bases, positions fo closing pairs, connections to stems, %s and absolute
          //numbers for make up, assymetry measure, explicitly number of branches (already reported)
          printf("closed region [%d, %d]\n\n",begin, end);
          cout << "***" << Name << Num << "***\n";
          MultiConnections(false, i);
          cout << endl;
          
          //want to take into account stacked pairs of length 1, thus... this might be a stacked pair, too      
          if(Parent->type != stackloop)
            SingleStacked(i);
            
          cout << "******************************************************************************\n\n";
    
        }
      else if(type==pseudo) //we must have a pseudo region, ie something that contains pseudoknots
        {
          printf("closed region [%d, %d]\n\n",begin, end);
          cout << "***" << Name << " " << Num << "***\n";
          PseudoType();
          Input->tot_bands+=NumberOfBands;
          cout << endl << "******************************************************************************\n\n";
        }
    }
        

  if(i == -1) //there is no real closed region characterized as external loop, thus...
    //external loop ourselves, this would have to be the outer loop
    {
      Input->tot_external++;
      //need: length of free bases on both 5' and 3' ends, plus everything as for multi
      cout << "\n***external loop " << Input->tot_external << "***\n";
      MultiConnections(true, i);
      cout << "\n******************************************************************************\n\n";
    }

  
  Loop * L = RightChild;

  if (L)
    {
      do 
        {
          L->Print(i+1);
          L = L->LeftSibling;
        } while (L != NULL);
    };
    
  if (i==-1)
    {
      cout << "##########################################################################################\n\n----------TOTALS:----------\n\n";
      //print the total number of pseudoknots and their locations
      cout << "Pseudoknot total " << Input->tot_pseudo << endl;
      if(Input->tot_pseudo > 0)
        {
          cout << "Pseudoknot coordinates:\n";
          //go through left and right and extract the end points
          for(int g=0; g< Input->tot_pseudo; g++)
            {
              cout << "\t" << g+1 << ") " << Input->left[g] << "-" << Input->right[g] << endl;   
            }
        }
      //print the total numbers of bands, and loop types
      cout << endl << "Bands (in pseudoknots) total " << Input->tot_bands << endl
        /*
          << "Pseudo Stem total " << Input->tot_pstems << endl
          << "Pseudo Hairpin total " << Input->tot_phairpins << endl
          << "Pseudo Bulge total " << Input->tot_pbulges << endl
          << "Pseudo Internal loop total " << Input->tot_pinter << endl
          << "Pseudo Multi loop total " << Input->tot_pmulti << endl
        */
           << "Base pairs (found in pseudo stems) total " << Input->tot_pstacks << endl
           << "Stem total " << Input->tot_stems << endl
           << "Base pairs (found in regular stems) total " << Input->tot_stacks  << endl
           << "Hairpin loop total " << Input->tot_hairpins << endl
           << "Bulge total " << Input->tot_bulges << endl
           << "Internal loop total " << Input->tot_inter << endl
           << "Multi loop total " << Input->tot_multi << endl
           << "External loop total " << Input->tot_external << endl
           << "Domain total " << Input->tot_domains << endl << endl;
          
      if(Input->tot_hairpins)
        {
          cout << "Hairpin Length Max: " << Input->max_h << endl
               << "Hairpin Length Min: " << Input->min_h << endl << "Hairpin Length Avg: ";
               printf("%.2f\n", (float) (Input->avg_h)/(Input->tot_hairpins));
        }
           
      if(Input->tot_inter)
        {
          cout << "Internal Length Max: " << Input->max_i << endl
               << "Internal Length Min: " << Input->min_i << endl << "Internal Length Avg: ";
               printf("%.2f\n", (float) (Input->avg_i)/(Input->num_length_i));
        }
           
      if(Input->tot_bulges)
        {
          cout << "Bulge Length Max: " << Input->max_b << endl
               << "Bulge Length Min: " << Input->min_b << endl << "Bulge Length Avg: ";
               printf("%.2f\n", (float) (Input->avg_b)/(Input->num_length_b));
        }
          
      if(Input->tot_multi)
        { 
          cout << "Multi Length Max: " << Input->max_m << endl
               << "Multi Length Min: " << Input->min_m << endl << "Multi Length Avg: ";
               printf("%.2f\n", (float) (Input->avg_m)/(Input->num_length_m));
        }
         
      if(Input->tot_stems)
        {
          cout << "Stem Length Max: " << Input->max_stem << endl
               << "Stem Length Min: " << Input->min_stem << endl << "Stem Length Avg: ";
               printf("%.2f\n", (float) (Input->avg_stem)/(Input->tot_stems));
        }
           
      if(Input->tot_bands)
        {
          cout << "Band Length Max: " << Input->max_band << endl
               << "Band Length Min: " << Input->min_band << endl << "Band Length Avg: ";
               printf("%.2f\n", (float) (Input->avg_band)/(Input->tot_bands));
        }
           
      if(Input->tot_stems)
        {  
          cout << "\nTotal numbers and percentages for bases from all stems in the molecule:\n";
          for(int g=0; g<MaxBases-1; g++)
            {
              if(Input->tot_bases[g])
              {
                cout << Input->tot_bases[g] << " " << Input->basesChar[g] << "'s, ";
                printf("%.2f%%\n", (float) (Input->tot_bases[g]*100)/(Input->tot_Bases));
                }
            }
            if(Input->tot_bases[MaxBases-1])
            printf("%d other bases, %.2f%%\n", Input->tot_bases[MaxBases-1], (float) (Input->tot_bases[MaxBases-1]*100)/(Input->tot_Bases));
          cout << "Total numbers and percentages for base pairs from all stems in the molecule:\n";
     
          for(int g=0; g<MaxPairs; g++)
            {
              if(Input->tot_pairs[g])
              {
                cout << Input->tot_pairs[g] << " "<< Input->pairsChar[g] <<"'s, ";
                 printf("%.2f%%\n", (float) (Input->tot_pairs[g]*100)/(Input->tot_Pairs));
                 }
            }
        }
        
      if(Input->tot_pseudo)
        {
          cout << "\nTotal numbers and percentages for bases from all pseudo stems in the molecule:\n";
          for(int g=0; g<MaxBases-1; g++)
            {
              if(Input->tot_pbases[g])
              {
                cout << Input->tot_pbases[g] << " " << Input->basesChar[g] << "'s, ";
                printf("%.2f%%\n", (float) (Input->tot_pbases[g]*100)/(Input->tot_pBases));
                }
            }
            if(Input->tot_pbases[MaxBases-1])
            printf("%d other bases, %.2f%%\n", Input->tot_pbases[MaxBases-1], (float) (Input->tot_pbases[MaxBases-1]*100)/(Input->tot_pBases));
          cout << "Total numbers and percentages for base pairs from all pseudo stems in the molecule:\n";
     
          for(int g=0; g<MaxPairs; g++)
            {
              if(Input->tot_ppairs[g])
              {
                cout << Input->tot_ppairs[g] << " "<< Input->pairsChar[g] <<"'s, ";
                printf("%.2f%%\n", (float) (Input->tot_ppairs[g]*100)/(Input->tot_pPairs));
                }
            }
        }        
        
      if(Input->tot_hairpins)
        {
          cout << "\nTotal numbers and percentages for free bases from all hairpins in the molecule:\n";
          for(int g=0; g<MaxBases-1; g++)
            {
              if(Input->fbh[g])
              {
                cout << Input->fbh[g] << " " << Input->basesChar[g] << "'s, ";
                printf("%.2f%%\n", (float) (Input->fbh[g]*100)/(Input->tot_fbh));
                }
            }
            if(Input->fbh[MaxBases-1])
            printf("%d other bases, %.2f%%\n", Input->fbh[MaxBases-1], (float) (Input->fbh[MaxBases-1]*100)/(Input->tot_fbh));
        
          cout << "Total numbers and percentages for closing base pairs from all hairpins in the molecule:\n";
          for(int g=0; g<MaxPairs; g++)
            {
              if(Input->cbph[g])
              {
                cout << Input->cbph[g] << " " << Input->pairsChar[g] << "'s, ";
                printf("%.2f%%\n", (float) (Input->cbph[g]*100)/(Input->tot_cbph));
              }
            }
        }
        
        
      if(Input->tot_bulges)
        {
          cout << "\nTotal numbers and percentages for free bases from all bulges in the molecule:\n";
          for(int g=0; g<MaxBases-1; g++)
            {
              if(Input->fbb[g])
              {
                cout << Input->fbb[g] << " " << Input->basesChar[g] << "'s, ";
                printf("%.2f%%\n", (float) (Input->fbb[g]*100)/(Input->tot_fbb));
                }
            }
            if(Input->fbb[MaxBases-1])
            printf("%d other bases, %.2f%%\n", Input->fbb[MaxBases-1], (float) (Input->fbb[MaxBases-1]*100)/(Input->tot_fbb));
        
        
          cout << "Total numbers and percentages for closing base pairs from all bugles in the molecule:\n";
          for(int g=0; g<MaxPairs; g++)
            {
              if(Input->cbpb[g])
              {
                cout << Input->cbpb[g] << " " << Input->pairsChar[g] << "'s, ";
                printf("%.2f%%\n", (float) (Input->cbpb[g]*100)/(Input->tot_cbpb));
              }
            }
        }
        

      if(Input->tot_inter)
        {
          cout << "\nTotal numbers and percentages for free bases from all internal loops in the molecule:\n";
          for(int g=0; g<MaxBases-1; g++)
            {
              if(Input->fbi[g])
              {
                cout << Input->fbi[g] << " " << Input->basesChar[g] << "'s, ";
                printf("%.2f%%\n", (float) (Input->fbi[g]*100)/(Input->tot_fbi));
                }
            }
            if(Input->fbi[MaxBases-1])
            printf("%d other bases, %.2f%%\n", Input->fbi[MaxBases-1], (float) (Input->fbi[MaxBases-1]*100)/(Input->tot_fbi));
        
                  cout << "Total numbers and percentages for closing base pairs from all internal loops in the molecule:\n";
          for(int g=0; g<MaxPairs; g++)
            {
              if(Input->cbpi[g])
              {
                cout << Input->cbpi[g] << " " << Input->pairsChar[g] << "'s, ";
                printf("%.2f%%\n", (float) (Input->cbpi[g]*100)/(Input->tot_cbpi));
                }
            }
        }
        
        
      if(Input->tot_multi)
        {
          cout << "\nTotal numbers and percentages for free bases from all multi loops in the molecule:\n";
          for(int g=0; g<MaxBases-1; g++)
            {
              if(Input->fbm[g])
              {
                cout << Input->fbm[g] << " " << Input->basesChar[g] << "'s, ";
                printf("%.2f%%\n", (float) (Input->fbm[g]*100)/(Input->tot_fbm));
                }
            }
            if(Input->fbm[MaxBases-1])
            printf("%d other bases, %.2f%%\n", Input->fbm[MaxBases-1], (float) (Input->fbm[MaxBases-1]*100)/(Input->tot_fbm));
        
                      cout << "Total numbers and percentages for closing base pairs from all multi loops in the molecule:\n";
          for(int g=0; g<MaxPairs; g++)
            {
              if(Input->cbpm[g])
              {
                cout << Input->cbpm[g] << " " << Input->pairsChar[g] << "'s, ";
                printf("%.2f%%\n", (float) (Input->cbpm[g]*100)/(Input->tot_cbpm));
              }
            }
        }
            
        
      if(Input->tot_fbe)
        {
          cout << "\nTotal numbers and percentages for free bases from the external loop of the molecule:\n";
          for(int g=0; g<MaxBases-1; g++)
            {
              if(Input->fbe[g])
              {
                cout << Input->fbe[g] << " " << Input->basesChar[g] << "'s, ";
                printf("%.2f%%\n", (float) (Input->fbe[g]*100)/(Input->tot_fbe));
                }
            }
            if(Input->fbe[MaxBases-1])
            printf("%d other bases, %.2f%%\n", Input->fbe[MaxBases-1], (float) (Input->fbe[MaxBases-1]*100)/(Input->tot_fbe));
        
            
                      cout << "Total numbers and percentages for closing base pairs from the external loop of the molecule:\n";
          for(int g=0; g<MaxPairs; g++)
            {
              if(Input->cbpe[g])
              {
                cout << Input->cbpe[g] << " " << Input->pairsChar[g] << "'s, "; 
                printf("%.2f%%\n", (float) (Input->cbpe[g]*100)/(Input->tot_cbpe));
                
              }
            }
        }
        
        
      if(Input->tot_fbwb)
        {
          cout << "\nTotal numbers and percentages for free bases found within bands:\n";
          for(int g=0; g<MaxBases-1; g++)
            {
              if(Input->fbwb[g])
              {
                cout << Input->fbwb[g] << " " << Input->basesChar[g] << "'s, ";
                 printf("%.2f%%\n", (float) (Input->fbwb[g]*100)/(Input->tot_fbwb));
              }
            }
            if(Input->fbwb[MaxBases-1])
            printf("%d other bases, %.2f%%\n", Input->fbwb[MaxBases-1], (float) (Input->fbwb[MaxBases-1]*100)/(Input->tot_fbwb));
        }
        
    }
}                                                            


//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS


Loop * Loop::StemConnections(int index)
{
  Loop *STK = RightChild; // this won't be null, since otherwise we wouldn't have a stack on our hands to begin with
  int brr=2;

  while((STK->RightChild != NULL)&&((begin+brr == STK->RightChild->begin)&&(end-brr == STK->RightChild->end)))
    {
      STK = STK->RightChild;
      brr++;
    }
  int length = ((STK->begin)-begin)+1;
  cout << "Coordinates: " << begin << "-" << end << ", " << STK->begin << "-" << STK->end << endl
       << "Length (in paired bases): " << length << endl << "Connections: ";
       
  if(Input->avg_stem==0)
    {
      Input->max_stem=length;
      Input->min_stem=length;
      Input->avg_stem=length;
    }
  else
    {
      Input->max_stem=max(Input->max_stem,length);
      Input->min_stem=min(Input->min_stem,length);
      Input->avg_stem+=length;
    }

  if(index==0)
    {
      cout << "external loop 1"; 
    }
  else
    cout << Parent->Name << Parent->Num; 
    
  cout << ", " << STK->Name << STK->Num << endl;
   
  return STK;
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

int Loop::standard_base(char nucleotide)
{
  switch(nucleotide)
    {
    case 'a': case 'A': case 'c': case 'C': case 'g': 
    case 'G': case 'u': case 'U': case 't': case 'T':
      return 1;
    default: return 0;        
    }        
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

int Loop::nuc_to_int(char nucleotide)
{
  switch(nucleotide)
    {
    case 'a': case 'A': return 0;
    case 'c': case 'C': return 1;
    case 'g': case 'G': return 2;
    default : return 3;
    }
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

double Loop::StemFreeEnergy(int pos1, int pos2)
{
  //first we need to extract the two char arrays that we will be manipulating
  // ie char *seq1, char *seq2 are needed
  //with first sequence going from "begin" to STK->begin
  //and the second- from STK->end to "end"
  int seq_size = (pos2-pos1)+1;
  char seq1[seq_size+1];
  char seq2[seq_size+1];
  int s1, s2;
  s1=pos1;
  s2=Input->BasePair(pos2);
  
  
  //this is for printing out seq1 and seq2, for debugging purposes.
  int z=0;
  for(z=0; z<seq_size; z++)
    {
      seq1[z] = Input->CSequence[s1+z];
      seq2[z] = Input->CSequence[s2+z];
    }
  seq1[z]='\0';
  seq2[z]='\0';
  
  //now we can proceed with the actual calculation, provided by Mirela
  int stacking_en[4][4][4][4] = {{{{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0}},
                                  {{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0}},
                                  {{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0}},
                                  {{0, 0, 0, -90},{0, 0, -220, 0},{0, -210, 0, -60},{-110, 0, -140, 0}}},
                                 {{{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0}},
                                  {{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0}},
                                  {{0, 0, 0, -210},{0, 0, -330, 0},{0, -240, 0, -140},{-210, 0, -210, 0}},
                                  {{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0}}},
                                 {{{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0}},
                                  {{0, 0, 0, -240},{0, 0, -340, 0},{0, -330, 0, -150},{-220, 0, -250, 0}},
                                  {{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0}},
                                  {{0, 0, 0, -130},{0, 0, -250, 0},{0, -210, 0, -50},{-140, 0, 130, 0}}},
                                 {{{0, 0, 0, -130},{0, 0, -240, 0},{0, -210, 0, -100},{-90, 0, -130, 0}},
                                  {{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0}},
                                  {{0, 0, 0, -100},{0, 0, -150, 0},{0, -140, 0, 30},{-60, 0, -50, 0}},
                                  {{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0}}}};

  int length, i, energy;
  length = strlen (seq1);
  energy = 0;
  for (i=0; i < length-1; i++)   // there are len-1 stacked pairs
    {
      energy += stacking_en [nuc_to_int (seq1[i])]
        [nuc_to_int (seq2[length-i-1])]
        [nuc_to_int (seq1[i+1])]
        [nuc_to_int (seq2[length-i-2])]; 
    }

  return energy/100.0;
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::StemContext(int pos1, int cur_pair)
{
  int pos2=Input->BasePair(pos1);
  /*  
      non-canonical base pair AG at position 10-50
      5' (i-2) base: A, 3'(j+2) base: G, paired = true
      5' (i-1) base: C, 3'(j+1) base: G, paired = true
      5' (i+1) base: A, 3'(j-1) base: G, paired = false
      5' (i+2) base: A, 3'(j-2) base: G, paired = false

      if between i and j, there is:
      one free base -> consider it as i+1
      two -> then i+1 and j-1
      three -> then i+1, i+2, and j-1  
  */

  cout << "Non-canonical base pair " << Input->CSequence[pos1] << Input->CSequence[pos2] << " at position " << pos1 << "-" << pos2 << endl;

  string s1, s2, s3, s4, s5, s6, s7;
  s1="5' (i";
  s2=") base: ";
  s3=", 3'(j";
  s4=", paired = ";
  s5="none";
  s6="false\n";
  s7="true\n";

  //this is on the outside of our arc
  cout << s1 << "-2" << s2;
  if((pos1-2)<1)
    cout << s5;
  else
    cout << Input->CSequence[pos1-2];
  cout << s3 << "+2" << s2;
  if((pos2+2)>Input->Size)
    cout << s5;
  else
    cout << Input->CSequence[pos2+2];
  cout << s4;
  //check if paired
  if((Input->BasePair(pos1-2)!=pos2+2)||(pos2+2==0))
    cout << s6;
  else
    cout << s7;

  cout << s1 << "-1" << s2;
  if((pos1-1)<1)
    cout << s5;
  else
    cout << Input->CSequence[pos1-1];
  cout << s3 << "+1" << s2;
  if((pos2+1)>Input->Size)
    cout << s5;
  else
    cout << Input->CSequence[pos2+1];
  cout << s4;
  //check if paired
  if((Input->BasePair(pos1-1)!=pos2+1)||(pos2+1==0))
    cout << s6;
  else
    cout << s7;

  //this is on the inside of our arc
  cout << s1 << "+1" << s2
       << Input->CSequence[pos1+1]   //no need to check this as i and j won't be consequent
       << s3 << "-1" << s2;
  if((pos2-1)>(pos1+1))
    cout << Input->CSequence[pos2-1];
  else
    cout << s5;
  cout << s4;
  //check if paired
  if(((Input->BasePair(pos1+1)!=pos2-1)||(pos2-1==0))||(pos1==pos2-1))
    cout << s6;
  else
    cout << s7;

  cout << s1 << "+2" << s2;
  if((pos1+2)<(pos2-1))
    cout << Input->CSequence[pos1+2];
  else
    cout << s5;
  cout << s3 << "-2" << s2;
  if((pos2-2)>(pos1+2))
    cout << Input->CSequence[pos2-2];
  else
    cout << s5;
  cout << s4;
  //check if paired
  if(((Input->BasePair(pos1+2)!=pos2-2)||(pos2-2==0))||(pos1==pos2-2))
    cout << s6;
  else
    cout << s7;
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::Numbers(int pos1, int pos3, bool nonband)
{
  if(pos1>pos3)
    return;

  int tot1, tot2, k, a, b;
  int bases[5];
  int pairs[11];
  
  tot1=tot2=k=0;
  
  for(int i=0; i<5; i++)
      bases[i]=0;
  
  for(int i=0; i<11; i++)
      pairs[i]=0;
  

  for(int z=0; z<(pos3-pos1+1); z++)
    {
      k = BPRec(pos1+z,a,b);
          
      tot1 += 2;
      tot2++;
      pairs[k]++;
      bases[a]++;
      bases[b]++;
            
      //now check for neigbours, if our base pair is NC
      if(k<7)
        StemContext(pos1+z, k);
          
      if(nonband)
        {
          Input->tot_Bases +=2;
          Input->tot_Pairs++;
          Input->tot_pairs[k]++;
          Input->tot_bases[a]++;
          Input->tot_bases[b]++;
        }
      else
        {
          Input->tot_pBases +=2;
          Input->tot_pPairs++;
          Input->tot_ppairs[k]++;
          Input->tot_pbases[a]++;
          Input->tot_pbases[b]++;
        }
    }
    

  cout << "Total numbers, percentages of bases involved in this ";
  if(nonband)
    cout << "stem:\n";
  else
    cout << "band stem:\n";

  for(int i=0; i<MaxBases-1; i++)
    {
      if(bases[i])
      printf("%d %c's, %.2f%%\n", bases[i], Input->basesChar[i], (float) (bases[i]*100)/tot1);
    }
    if(bases[MaxBases-1])
    printf("%d other bases, %.2f%%\n", bases[MaxBases-1], (float) (bases[MaxBases-1]*100)/tot1);

  cout << "Total numbers, percentages of basepairs involved in this ";
  if(nonband)
    cout << "stem:\n";
  else
    cout << "band stem:\n";
  
  for(int i=0; i<MaxPairs; i++)
    {
      if(pairs[i])
      {
        cout << pairs[i] << " "<< Input->pairsChar[i];
        printf("'s, %.2f%%\n", (float) (pairs[i]*100)/tot2);
      }
    } 
    
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

int Loop::BPRec(int pos1, int& c, int& d) //recognition of a base pair
{
  int pos2 = Input->BasePair(pos1);
  char a = Input->CSequence[pos1];
  char b = Input->CSequence[pos2];


  if(a==b)
    {
      if(a=='A') 
        {
        c=d=0;
          return 0;
        }
      else if(a=='C') 
        {
        c=d=2;
          return 2;
        }
      else if(a=='G') 
        {
        c=d=3;
          return 3;
        } 
      else if(a=='U') 
        {
        c=d=1;
          return 1;
        }
        else
        {
        c=d=4;
        return 10;
        }
    }
  else
    {
      if((a=='A' && b=='U')||(a=='U' && b=='A')) 
        {
        c=0;
        d=1;
          return 8;
        }
      else if((a=='C' && b=='G')||(a=='G' && b=='C')) 
        {
        c=2;
        d=3;
          return 9;
        }
      else if((a=='A' && b=='C')||(a=='C' && b=='A')) 
        {
        c=0;
        d=2;
          return 4;
        }
      else if((a=='A' && b=='G')||(a=='G' && b=='A')) 
        {
        c=0;
        d=3;
          return 5;
        }
      else if((a=='U' && b=='C')||(a=='C' && b=='U')) 
        {
        c=1;
        d=2;
          return 6;
        }
      else if((a=='G' && b=='U')||(a=='U' && b=='G')) 
        {
        c=1;
        d=3;
          return 7;
        }
        //no deal with other cases, where at least one base is not from the original four
        //first deal with a base pair that contains one normal base and another being the "other"
        else if((a=='A')||(b=='A'))
        {
        c=0;
        d=4;
        return 10;
        }
        else if((a=='U')||(b=='U'))
        {
        c=1;
        d=4;
        return 10;
        }
        else if((a=='C')||(b=='C'))
        {
        c=2;
        d=4;
        return 10;
        }
        else if((a=='G')||(b=='G'))
        {
        c=3;
        d=4;
        return 10;
        }
        else
        {
        c=d=4;
        return 10;
        }
    }
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::SingleStacked(int index)
{
  Input->tot_stems++;
  cout << "***stem " << Input->tot_stems << "***\n"
       << "Coordinates: " << begin << "-" << end << ", " << begin << "-" << end << endl
       << "Length (in paired bases): 1\n"
       << "Connections: ";
        
  if(index==0)
    cout << "external loop 1, ";
  else        
    {
      if(Parent->type!=stackloop)
        cout << Parent->Name << Parent->Num << ", ";
      else
        {
          Loop * PR = Parent;
          while((PR->Parent->type==stackloop)&&((PR->begin==PR->Parent->begin+1)&&(PR->end==PR->Parent->end-1)))
            {
              PR=PR->Parent;
            }
          cout << PR->Name << PR->Num << ", ";
        }
    }
         
  cout << Name << Num << endl << "Free Stem Energy: " << StemFreeEnergy(begin, begin) << endl;
  Numbers(begin, begin, true);
  cout << endl;
  
  if(Input->avg_stem==0)
    {
      Input->max_stem=1;
      Input->min_stem=1;
      Input->avg_stem=1;
    }
  else
    {
      Input->max_stem=max(Input->max_stem,1);
      Input->min_stem=min(Input->min_stem,1);
      Input->avg_stem++;
    }
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::HairpinConnections(int index)
{
  cout << "Coordinates: " << begin << "-" << end << endl
       << "Length (in unpaired bases): " << NumberOfUnpaird << endl
       << "Sequence: ";
  
  for(int i=begin; i<=end; i++)
    {
    cout << Input->CSequence[i]; 
    if((i!=begin)&&(i!=end))
      TypeFreeBase(i, false);
    }

  cout << endl << "Connections: ";
  
  if(index==0)
    cout << "external loop 1\n";
  else
    {
      if(Parent->type!=stackloop)
        cout << Parent->Name << Parent->Num << endl;
      else
        {
          Loop* PR = Parent;
          while((PR->Parent->type==stackloop)&&((PR->begin==PR->Parent->begin+1)&&(PR->end==PR->Parent->end-1)))
            {
              PR=PR->Parent;
            }
          cout << PR->Name << PR->Num << endl;
        }
    }
 

  IncFreeBase();
  
  int a, b;

  ClosingBasePair(BPRec(begin,a,b),false);
  IncCBP();


  //do stats on length
  if(Input->avg_h==0)
    {
      Input->max_h=NumberOfUnpaird;
      Input->min_h=NumberOfUnpaird;
      Input->avg_h=NumberOfUnpaird;
    }
  else
    {
      Input->max_h=max(Input->max_h,NumberOfUnpaird);
      Input->min_h=min(Input->min_h,NumberOfUnpaird);
      Input->avg_h+=NumberOfUnpaird;
    }
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS
            
void Loop::HairpinType()
{

  //open up the file with motifs, traverse it, and check if our hairpin fits one of the motifs
  ifstream fileStr;
  fileStr.open(MotifFile);
  char junk[256];

  //get the comments marked by '#' at the beginning of the line out
  //remove comments from file
  while(fileStr.peek() == '#')
    {
      fileStr.getline(junk, MAXLINE);
    }

  char mot;
  char hair;
  string motif = "";
  bool match = false;
  bool stillmatch = true;
  //we have our begin and end implicitly present
  int i=begin+1;

  //********************************************   
  //could have the following characters in a motif:
  // A-adenine
  // U-uracil
  // G-guanine
  // C-cytosine
  // N-any of A,U,G or C
  // R-purine: A or G
  // Y-pyrimidine: U or C
  //********************************************
  //the motifs in a file could be separated by space, comma, tab, vertical tab, newline char, or carrige return
    
  //what the code right now is supposed to do:
  //read in a motif per time, as we are doing that we compare each character with the character of
  //the csequence occupied by our hairpin, if it "stillmatches" then we keep going with this motif
  //if not then we go to next motif, we also control for length, if motif is too long, we forget about it
  //as we are traversing each motif, as long as it matches, we push it onto our "motif" string, so we could
  //later on, if our hairpin matches this motif, say exactly which motif it matches.
      
  while((!fileStr.eof()) && fileStr.good())
    {
      mot = fileStr.get();
      if(fileStr.eof())
        break;
      if(((mot != ' ')&&(mot != ','))&&(((mot != '\t')&&(mot != '\n'))&&((mot != '\r')&&(mot != '\v'))))
        {
          while((((mot!=' ')&&(mot!=','))&&((mot!='\t')&&(mot!='\n'))&&((mot!='\r')&&(mot!='\v')))&&(stillmatch)) 
            {
              if(i<end)
                {
                  hair = Input->CSequence[i];
                  //cout << "We are on this hair now: " << hair << endl;
                  if((mot=='A')&&(hair==mot))
                    {
                      //cout << "Just matched A\n";
                      motif.append(1, mot);
                      match=true;
                    }
                  else if((mot=='U')&&(hair==mot))
                    {
                      //cout << "Just matched U\n";
                      motif.append(1, mot);
                      match=true;
                    }
                  else if((mot=='G')&&(hair==mot))
                    {
                      //cout << "Just matched G\n";
                      motif.append(1, mot);
                      match=true;
                    }
                  else if((mot=='C')&&(hair==mot))
                    {
                      //cout << "Just matched C\n";
                      motif.append(1, mot);
                      match=true;
                    }
                  else if(mot=='N')
                    {
                      //cout << "Just matched N\n";
                      motif.append(1, mot);
                      match=true;
                    }
                  else if((mot=='R')&&((hair=='A')||(hair=='G')))
                    {
                      //cout << "Just matched R\n";
                      motif.append(1, mot);
                      match=true;
                    }
                  else if((mot=='Y')&&((hair=='U')||(hair=='C')))
                    {
                      //cout << "Just matched Y\n";
                      motif.append(1, mot);
                      match=true;
                    }
                  else
                    {
                      //cout << "Just unmatched \n";
                      stillmatch=false;
                      match=false;
                    }
                  i++;
                }
              else
                {
                  stillmatch=false;
                  match=false;
                }
              mot = fileStr.get();
            }
          if((match))//&&(NumberOfUnpaird==motif.size()))
            {
              cout << "Matches motif: " << motif << endl;
              fileStr.close();
              return;
            }
          i=begin+1;
          stillmatch=true;
          match=false;
          motif.clear();
          while((((fileStr.peek()!=' ')&&(fileStr.peek()!=','))&&((fileStr.peek()!='\t')&&(fileStr.peek()!='\n'))&&((fileStr.peek()!='\r')&&(fileStr.peek()!='\v')))&&(stillmatch))
            {
              mot = fileStr.get();
              if(fileStr.eof())
                return;
            }
        }
    }
    
  cout << "No matches were found for any of the motifs.\n";
  fileStr.close();
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::MultiConnections(bool external, int index)
{
  cout << "Coordinates: ";
   
  if(!external)
    cout << begin << "-" << end << ", ";

  vector<int> coord;
  if(!external)
    coord.push_back(end);

  //go through the children
  Loop * Child = RightChild;
  while(Child!=NULL)
    {
      coord.push_back(Child->end);
      coord.push_back(Child->begin);
      cout << Child->begin << "-" << Child->end;
      if(Child->LeftSibling)
        cout << ", ";
      Child=Child->LeftSibling; 
    }

  if(!external)
    coord.push_back(begin);

  /*
    cout << "\n\nCOORD: ";
    for(int g=0; g<coord.size(); g++)
    {
    cout << coord[g] << ", ";
    }
  */

  cout << endl << "Lengths (in unpaired bases): ";
  int size=coord.size();  
  int length=0;
  int num_stems=0;

  if(!external)
    {
      for(int i=1; i<size; i++)
        {
          for(int j=coord[i]+1; j<coord[i-1]; j++)
            {
              if(Input->Sequence[j] <= 0)
                {
                  length++;
                  TypeFreeBase(j, external);
                }
            }
          cout << length;
          num_stems++;
          lengths[i/2]=length;
          length=0;
          i++;
          if(i<size-1)
            cout << ", ";
        }
      cout << endl;
    }
  else
    {
      cout << "\n\t on 5\' end: ";
      for(int i=1; i<coord[size-1]; i++)
        {
          if(Input->Sequence[i] <= 0)
            {
              length++;
              TypeFreeBase(i, external);
            }
        }
      cout << length;
      lengths[0]=length;
      length=0;
      cout << "\n\t on 3\' end: ";
      for(int i=coord[0]+1; i<=Input->Size; i++)
        {
          if(Input->Sequence[i] <= 0)
            {
              length++;
              TypeFreeBase(i, external);  
            }
        }
      cout << length;
      lengths[1]=length;
      num_stems++;

      if(size>2)
        {
          cout <<"\n\t and the rest are: ";
          length=0;

          for(int i=2; i<size-1; i++)
            {
              for(int j=coord[i]+1; j<coord[i-1]; j++)
                {
                  if(Input->Sequence[j] <= 0)
                    {
                      length++;
                      TypeFreeBase(j, external);
                    }
                }
              i++;
              cout << length;
              lengths[(i+2)/2]=length;
              num_stems++;
              length=0;
              if(i<size-1)
                cout << ", ";
            }
        }
      cout << endl;
    }

  if((type==multi)||(external))
    {
      cout << "Number of branches: " << num_stems << endl;
      size_lengths=num_stems;
      if(external)
      {
        Input->tot_domains=num_stems;
        size_lengths++;
      }
      MultiAsymmetry();
    }
  else //if((type==interior)||()
    {
      size_lengths=2;
      MultiAsymmetry();
    }

  cout << "Connections: ";
  if(!external)
    {
      if(index==0)
        cout << "external loop 1, ";
      else
        {
          if(Parent->type!=stackloop)
            cout << Parent->Name << Parent->Num << ", ";
          else
            {
              Loop* PR = Parent;
              while((PR->Parent->type==stackloop)&&((PR->begin==PR->Parent->begin+1)&&(PR->end==PR->Parent->end-1)))
                {
                  PR=PR->Parent;
                }
              cout << PR->Name << PR->Num << ", ";
            }
        }
    }
 
  Child=RightChild;

  while(Child->LeftSibling)
    {
      cout << Child->Name << Child->Num << ", ";
      Child=Child->LeftSibling;
    }

  cout << Child->Name << Child->Num << endl;
  
  /*
  //FOR DEBUGGING
  cout << "THE LENGTH ARRAY: ";
  for (int i=0; i<size_lengths; i++)
  cout << lengths[i] << ",";
  cout << endl << "The size is: " << size_lengths << endl;
  */
  
  
  //do stats on lengths
  if(type==multi)
    {
      if(Input->num_length_m==0)
        {
          Input->avg_m=lengths[0];
          Input->max_m=lengths[0];
          Input->min_m=lengths[0];
          for(int i=1; i<size_lengths; i++)
            {
              Input->avg_m+=lengths[i];
              Input->max_m=max(Input->max_m,lengths[i]);
              Input->min_m=min(Input->min_m,lengths[i]);
            }
          Input->num_length_m=size_lengths;
        }
      else
        {
          for(int i=0; i<size_lengths; i++)
            {
              Input->avg_m+=lengths[i];
              Input->max_m=max(Input->max_m,lengths[i]);
              Input->min_m=min(Input->min_m,lengths[i]);
            }
          Input->num_length_m+=size_lengths;
        }
    }
  else if(type==interior)
    {
      if(Input->num_length_i==0)
        {
          Input->avg_i=lengths[0];
          Input->max_i=lengths[0];
          Input->min_i=lengths[0];
          for(int i=1; i<size_lengths; i++)
            {
              Input->avg_i+=lengths[i];
              Input->max_i=max(Input->max_i,lengths[i]);
              Input->min_i=min(Input->min_i,lengths[i]);
            }
          Input->num_length_i=size_lengths;  
        }
      else
        {
          for(int i=0; i<size_lengths; i++)
            {
              Input->avg_i+=lengths[i];
              Input->max_i=max(Input->max_i,lengths[i]);
              Input->min_i=min(Input->min_i,lengths[i]);
            }
          Input->num_length_i+=size_lengths;
        }
    }
  else if(type==bulge)
    {
      if(Input->num_length_b==0)
        {
          Input->avg_b=lengths[0];
          Input->max_b=lengths[0];
          Input->min_b=lengths[0];
          for(int i=1; i<size_lengths; i++)
            {
              Input->avg_b+=lengths[i];
              Input->max_b=max(Input->max_b,lengths[i]);
              Input->min_b=min(Input->min_b,lengths[i]);
            }
          Input->num_length_b=size_lengths;
        }
      else
        {
          for(int i=0; i<size_lengths; i++)
            {
              Input->avg_b+=lengths[i];
              Input->max_b=max(Input->max_b,lengths[i]);
              Input->min_b=min(Input->min_b,lengths[i]);
            }
          Input->num_length_b+=size_lengths;
        }
    }
    
  IncFreeBase();
      
  //now tell exactly what the closing pairs are
  int k, a, b;
      
  if(external)
    {
      for(int i=0; i<size; i++)
        {
          k = BPRec(coord[i],a,b);
          ClosingBasePair(k,external);
          i++;
        }
    }
  else
    {
      k = BPRec(coord[0],a,b);
      ClosingBasePair(k,external);
      
      for(int i=1; i<size-1; i++)
        {
          k = BPRec(coord[i],a,b);
          ClosingBasePair(k,external);
          i++;
        }
    }
      
  IncCBP();
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::MultiAsymmetry()
{
  if(size_lengths > 1)
    {
      cout << "Asymmetry:\n";
      if(size_lengths==2)
        {
          printf("\tAbsolute Difference: %.2f\n", (float) abs(lengths[0]-lengths[1]));
          if(lengths[0]!=0)
            {
              if(lengths[1]!=0)
                printf("\tRatio: %.2f\n", (float) max(lengths[0]/lengths[1], lengths[1]/lengths[0]));
              else
                cout << "\tRatio: 0\n";
            }
          else 
            {
              if(lengths[1]==0)
                cout << "\tRatio: unattainable (all lengths are zero, can't divide by zero)\n";
              else
                cout << "\tRatio: 0\n";
            }
        }
      else
        {
          float maxAD, avgAD, minAD, maxR, avgR, minR;
          maxAD=avgAD=maxR=avgR=0;
          minAD= (float) abs(lengths[0]-lengths[1]);
          minR= (float) lengths[0];
          int avg_size=0;
          int avgAD_size=0;
          
          /*
          for(int i=0; i<size_lengths;i++)
            {
              if(lengths[i]!=0)
                {
                  minR=lengths[0]/lengths[i];
                  break;
                }
            }
          */

          for(int i=0; i<size_lengths; i++)
            {
              for(int j=0; j<size_lengths; j++)
                {
                  if((i!=j)&&(lengths[j]!=0))
                        {
                          maxR = (float) max(maxR, (float) lengths[i]/lengths[j]);
                          minR = (float) min(minR, (float) lengths[i]/lengths[j]);
                          avgR += (float) lengths[i]/lengths[j];
                          avg_size++;
                        }
                }
            }
            
            for(int i=0; i<size_lengths-1; i++)
            {
              for(int j=i+1; j<size_lengths; j++)
                {
                          maxAD = (float) max(maxAD, (float) abs(lengths[i]-lengths[j]));
                          minAD = (float) min(minAD, (float) abs(lengths[i]-lengths[j]));
                          avgAD += (float) abs(lengths[i]-lengths[j]); 
                          avgAD_size++;        
                }
            }
            
            
            
          if(avg_size==0)
            avg_size=avgAD_size=1;
          printf("\tAbsolute Difference:\n\t\tmax: %.2f\n\t\tmin: %.2f\n\t\tavg: %.2f\n",(float) maxAD, (float) minAD, (float) avgAD/avgAD_size); 
          printf("\n\tRatio:\n\t\tmax: %.2f\n\t\tmin: %.2f\n\t\tavg: %.2f\n", (float) maxR, (float) minR, (float) avgR/(avg_size));

        }
    }
  //else if(size_lengths==1)
  //  cout << "This external loop can be viewed as a hairpin\n";
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::TypeFreeBase(int i, bool external)
{
  //figures out what type of free base we are looking at and increments the local array, plus the total variable
  bool incremented=false;
  if(!external)
    {
      for(int j=0; j<MaxBases-1; j++)
        {
          if(Input->CSequence[i]==Input->basesChar[j])
          {
            fb[j]++;
            incremented=true;
          }
        }
        if(!incremented)
        fb[MaxBases-1]++;
      tot_fb++;
    }
  else
    {
      for(int j=0; j<MaxBases-1; j++)
        {
          if(Input->CSequence[i]==Input->basesChar[j])
          {
            Input->fbe[j]++;
            incremented=true;
          }
        }
        if(!incremented)
        Input->fbe[MaxBases-1]++;
      Input->tot_fbe++;
    }
}
     

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::IncFreeBase()
{
  //does the appropriate incrementation of the global array (in Input.h) of specific type according to the local array
  //plus the total global variable (in Input.h) gets incremented accordingly

  if(type==interior)
    {
      for(int i=0; i<MaxBases; i++)
        {
          if(fb[i]!=0)
            {
              Input->fbi[i]+=fb[i];
            }
        }
      Input->tot_fbi+=tot_fb;
    }
  else if(type==bulge)
    {
      for(int i=0; i<MaxBases; i++)
        {
          if(fb[i]!=0)
            {
              Input->fbb[i]+=fb[i];
            }
        }
      Input->tot_fbb+=tot_fb;
    }
  else if(type==hairpin)
    {
      for(int i=0; i<MaxBases; i++)
        {
          if(fb[i]!=0)
            {
              Input->fbh[i]+=fb[i];
            }
        }
      Input->tot_fbh+=tot_fb;
    }
  else if(type==multi)
    {
      for(int i=0; i<MaxBases; i++)
        {
          if(fb[i]!=0)
            {
              Input->fbm[i]+=fb[i];
            }
        }
      Input->tot_fbm+=tot_fb;
    }
        
  //do the printing out of #'s and %'s
  cout << "Total numbers and percentages of free bases found:\n";
  bool printed=false;
  if(type==external)
    {
      for(int i=0; i<MaxBases-1; i++)
        {
          if(Input->fbe[i])
            {
              printf("%d %c's, %.2f%%\n", Input->fbe[i], Input->basesChar[i], (float) (Input->fbe[i]*100)/Input->tot_fbe);
              printed=true;
            }
        }
        if(Input->fbe[MaxBases-1])
        {
        printf("%d other bases, %.2f%%\n", Input->fbe[MaxBases-1], (float) (Input->fbe[MaxBases-1]*100)/Input->tot_fbe);
        printed=true;
        }
      if(!printed)
        cout << "none\n";
    }
  else
    {
      for(int i=0; i<4; i++)
        {
          if(fb[i])
          {
            printf("%d %c's, %.2f%%\n", fb[i], Input->basesChar[i], (float) (fb[i]*100)/tot_fb);
            printed=true;
          }
        }
        if(fb[MaxBases-1])
        {
        printf("%d other bases, %.2f%%\n", fb[MaxBases-1], (float) (fb[MaxBases-1]*100)/tot_fb);
        printed=true;
        }
        if(!printed)
        cout << "none\n";
    }
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::ClosingBasePair(int k, bool external)
{
  if(external)
    {
      Input->tot_cbpe++;
      Input->cbpe[k]++;
    }
  else
    {
      tot_cbp++;
      cbp[k]++;
    }
}


//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::IncCBP()
{
  //do some global incrementing first
  if(type==hairpin)
    {
      Input->tot_cbph+=tot_cbp;
      for(int i=0; i<MaxPairs; i++)
        {
          if(cbp[i])
            Input->cbph[i]+=cbp[i];
        }
    }
  else if(type==interior)
    {
      Input->tot_cbpi+=tot_cbp;
      for(int i=0; i<MaxPairs; i++)
        {
          if(cbp[i])
            Input->cbpi[i]+=cbp[i];
        }
    }
  else if(type==bulge)
    {
      Input->tot_cbpb+=tot_cbp;
      for(int i=0; i<MaxPairs; i++)
        {
          if(cbp[i])
            Input->cbpb[i]+=cbp[i];
        }
    }
  else if(type==multi)
    {
      Input->tot_cbpm+=tot_cbp;
      for(int i=0; i<MaxPairs; i++)
        {
          if(cbp[i])
            Input->cbpm[i]+=cbp[i];
        }
    }
 

  //now let's do some printing
  cout << "Total numbers and percentages for closing base pairs:" << endl;
  if(type==external)
    {
      for(int i=0; i<MaxPairs; i++)
        {
          if(Input->cbpe[i])
          {
            cout << Input->cbpe[i] << " " << Input->pairsChar[i];
            printf("'s, %.2f%%\n", (float) (Input->cbpe[i]*100)/Input->tot_cbpe);
          }
        }
    }
  else
    {
      for(int i=0; i<MaxPairs; i++)
        {
          if(cbp[i])
          {
          cout << cbp[i] << " " << Input->pairsChar[i];
          printf("'s, %.2f%%\n", (float) (cbp[i]*100)/tot_cbp);
          }
        }     
    }
}
 
 
//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

void Loop::PseudoType()
{
  //state numbers for different children, paired bases involved in the band, pseudo type
  int size_bb=4*NumberOfBands;
  int bandbasepairs[NumberOfBands];
  int brem=bands_remove.size()/4;
  int arem=0; 
  
  int in_child_size=NumberOfBandChild*2;
  int in_children[in_child_size];
  int un_child_size=NumberOfUnBandChild*2;
  int un_children[un_child_size];
  int in,un;
  in=un=0;
  
  Loop * child = RightChild;
  
  for(int j=0; j<NumberOfChildren; j++)
    {
    if(child->nested==inBand)
    {
      in_children[in]=child->begin;
      in_children[in]=child->end;
      in++;
    }
    else if(child->nested==unBand)
    {
      un_children[un]=child->begin;
      un_children[un]=child->end;
      un++;
    }
      child=child->LeftSibling;
    }

    /*
    cout << "OUR CHILDREN array: ";
    for(int j=0; j<child_size; j++)
    {
    cout << children[j] << ",";
    }
    cout << endl;
    */
    
  
  cout << "Coordinates: " << begin << "-" << end << endl
       << "Number of children: " << NumberOfChildren << endl
       << "Number of un-band children: " << NumberOfUnBandChild << endl
       << "Number of in-band children: " << NumberOfBandChild << endl
       << "Number of unpaired bases: " << NumberOfUnpaird << endl
       << "Number of bands: " << NumberOfBands << endl
       << "Number of bands to be removed: " << brem << endl
       << "Coordinates of bands to be removed: ";

       for(int r=0; r<brem; r++)
       {
       cout << bands_remove[r*4] << "-" << bands_remove[r*4+3] << ", " << bands_remove[r*4+1] << "-" << bands_remove[r*4+2];
       //now go through one of the band borders and get the basepairs in the band
       for(int w=bands_remove[r*4]; w<=bands_remove[r*4+1]; w++)
         {
           if(Input->BasePair(w)>0)
             arem++;
         }
       if(r<brem-1)
       cout << "; ";
       }
       cout << endl << "Number of arcs to be removed: " << arem << endl << "______________________________\n";
          
  int i, tempi, z, g, y;
  i=tempi=0;
  int substems=1;
          
  for(int k=0; k<size_bb; k+=4)
    {
      printf("BAND %d\nCOORDINATES: %d-%d, %d-%d\n", (k/4)+1, bandborders[k],bandborders[k+3],bandborders[k+1],bandborders[k+2]);
      for(z=bandborders[k]; z<=bandborders[k+1]; z++)
        {
          g=Input->BasePair(z);
          if((g>=bandborders[k+2])&&(g<=bandborders[k+3])) //within a stem
            {
              Input->tot_pstacks++;
              i++;
              tempi++;
            }
          else 
            {
              //we might have encountered a free base (but this is only on the left end point of the band
              if((g<=0)&&(NotInChild(g,in_children,in_child_size)))
                {
                  tot_fb++;
                  for(int j=0; j<4; j++)
                    {
                      if(Input->CSequence[z]==Input->basesChar[j])
                        fb[j]++;
                    }
                }
          
              if(tempi) //new stem
                {
                  cout << "\n>Stem " << substems << endl
                       << "Coordinates: " << z-tempi << "-" << Input->BasePair(z-tempi) << ", " << z-1 << "-" << Input->BasePair(z-1) << endl
                       << "Length (in paired bases):" << tempi << endl
                       << "Free Stem Energy: " << StemFreeEnergy(z-tempi, z-1) << endl;
                  Numbers(z-tempi,z-1,false); //if our band consists of many sub stems
                  tempi=0;
                  substems++;
                } 
            }
        }
      if(tempi)
        {
          cout << "\n>Stem " << substems << endl
               << "Coordinates: " << z-tempi << "-" << Input->BasePair(z-tempi) << ", " << z-1 << "-" << Input->BasePair(z-1) << endl
               << "Length (in paired bases): " << tempi << endl
               << "Free Stem Energy: " << StemFreeEnergy(z-tempi, z-1) << endl;
          Numbers(z-tempi,z-1,false); //if last stem of the band has not been outputed yet
        }
      //deal with free bases that are on the right end point of the band
      for(y=bandborders[k+2]; y<=bandborders[k+3]; y++)
        {
          g=Input->BasePair(y);
          if((g<=0)&&(NotInChild(g,in_children,in_child_size)))
            {
              tot_fb++;
              for(int j=0; j<4; j++)
                {
                  if(Input->CSequence[y]==Input->basesChar[j])
                    fb[j]++;
                }
            }
        }
        
      cout << "\nLENGTH OF BAND " << (k/4)+1 << " (in paired bases): " << i << endl 
           << "NUMBER OF STEMS IN BAND " << (k/4)+1 << ": " << substems << endl 
           << "NUMBER OF FREE BASES INSIDE THE BAND: " << tot_fb << endl;
      if(tot_fb)
        {
          //go through the local array, do print outs, and increment the global array, clean up the local array 
          //s.t. it would be ready for next iteration of the loop
          for(int j=0; j<4; j++)
            {
              if(fb[j])
                {
                  cout << fb[j] << " " << Input->basesChar[j] << "'s, " << (fb[j]*100)/tot_fb << "%\n";
                  Input->fbwb[j]+=fb[j];
                  fb[j]=0;
                }
            }
          Input->tot_fbwb+=tot_fb;
        }
      cout  << "______________________________\n";
      bandbasepairs[k/4]=i;
      substems=1;
      tot_fb=0;
      
      if(Input->avg_band==0)
        {
          Input->max_band=i;
          Input->min_band=i;
          Input->avg_band=i;
        }
      else
        {
          Input->max_band=max(Input->max_band,i);
          Input->min_band=min(Input->min_band,i);
          Input->avg_band+=i;
        }
      tempi=i=0;
    }

    //do count of free bases in pseudo loops. 
    //first reorder the band borders (sort them) together with the ends/beginings of un-band children
    //then go in between all these "borders" and pick out the free bases->pseudo loops 
    
    
    
    
              
  cout << "Type: ";
  //now the fun part :) the types classification, so far only 9 types of 14 are present
  if((NumberOfBands==2)&&(Parent->type==external))
    {
      //count children under each band first
      int ch1, ch2;
      ch1=ch2=0;
      
      for(int j=0; j<un_child_size-1; j++)
      {
      if((un_children[j] > bandborders[1])&&(un_children[j+1] < bandborders[2]))
      ch1++;
      else if((un_children[j] > bandborders[5])&&(un_children[j+1] < bandborders[6]))
      ch2++;
      j++;
      }
        
            
      int maximum = max(ch1,ch2);
            
      if(maximum==0)
        cout << "E-H";
      else if(maximum==1)
        cout << "E-I";
      else 
        cout << "E-M";
            
      if(bandbasepairs[0]==bandbasepairs[1])
        cout << ", also ambiguous\n";
      else
        cout << endl;
    }
  else if(((NumberOfBands==3)&&(bandborders[3]<bandborders[8]))&&(((bandborders[1]<bandborders[4])&&(bandborders[5]<bandborders[2]))&&((bandborders[9]<bandborders[6])&&(bandborders[7]<bandborders[10]))))
    //3 bands in total AND band1 does not interact with band3 AND band2 has one end inside band1 AND band2 has other end inside band3
    {
      int ch1, ch3;
      ch1=ch3=0;
      
      for(int j=0; j<un_child_size-1; j++)
      {
      if((un_children[j] > bandborders[1])&&(un_children[j+1] < bandborders[2]))
      ch1++;
      else if((un_children[j] > bandborders[9])&&(un_children[j+1] < bandborders[10]))
      ch3++;
      j++;
      }
            
      if(ch1==0)
        {
          if(ch3==0)
            cout << "H-H";
          else if(ch3==1)
            cout << "H-I";
          else 
            cout << "H-M";
        }
      else if(ch1==1)
        {
          if(ch3==0)
            cout << "H-I";
          else if(ch3==1)
            cout << "I-I";
          else 
            cout << "I-M";
        }
      else 
        {
          if(ch3==0)
            cout << "H-M";
          else if(ch3==1)
            cout << "I-M";
          else 
            cout << "M-M";
        }
            
      if(bandbasepairs[1]>=(bandbasepairs[0]+bandbasepairs[2]))
        cout << ", also special\n";
      else
        cout << endl;
    }
  else
    cout << "unclassified\n";
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

bool Loop::NotInChild(int pos, int* children, int child_size)
{
  for(int i=0; i<child_size-1; i++)
    {
      if((pos>children[i])&&(pos<children[i+1]))
        return false;
      i++;
    }
  return true;
}

//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS


  
  
//SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS

// function added by Mirela - started on April 14, 2005


#define MAX(A, B)      ((A) > (B) ? (A) : (B))

Loop * getParent (Loop * loop, Loop *&child, int level, int &metLevel)
// level 1 = stop at the end of first stem
// level 2 = include external loop if it's at the end of the first stem, otherwise ignore
// returns NULL if the loop with this parent should be ignored
{
    Loop *parent;
    parent = loop->Parent;
    child = NULL;
    metLevel = 1;
    if (parent == NULL) 
    {
        cout << "Error, loop without parent\n";
        exit (1);
    }   
    if (parent->type != stackloop)
    {
        return NULL;     // return parent;    
    }
    // we have to get to the end of the stem
    while (parent->Parent->type == stackloop)
        parent = parent->Parent;
    if (level == 1)        
        return parent;                
    else if (level == 2)
    {
        if (parent->Parent->type == external)
        {
            child = parent;
            return parent->Parent;            
        }
        else if (parent->Parent->type == interior || parent->Parent->type == bulge)
        {
            return getParent (parent->Parent, child, level-1, metLevel);
        }
        else
        {
            metLevel = 0;
            return parent;
        }
    }    
    return NULL;
}

/////////////////////
Loop * getChild (Loop *loop, Loop *& previousChild)
// not used
{
    Loop *child;
    if (previousChild == NULL)
    {
        child = loop->RightChild;
        previousChild = child;
    }
    else
    {
        child = previousChild->LeftSibling;
        previousChild = child;
    }       
    if (child == NULL) return NULL;         
    while (child->type == stackloop)
        child = child->RightChild;
    // go back one level
    child = child->Parent;    
    return child;    
}


/////////////
Loop * getLeftmostChild (Loop *loop, Loop *& previousChild, int level, int &metLevel)
{
    Loop *child;
    metLevel = 1;
    child = loop->RightChild;
    if (child == previousChild)
        return NULL;
    // forbid children which are not continued by stems
    if (child->type != stackloop)
        return NULL;
    while (child->LeftSibling != previousChild)
    {
        child = child->LeftSibling;
        if (child->type != stackloop)
            return NULL;        
    }
    previousChild = child;
    // now go to the end of the stem
    while (child->type == stackloop)
        child = child->RightChild;
    child = child->Parent;    
    if (level == 1)
        return child;    
    else if (level == 2)
    {
        child = child->RightChild;
        if (child->type == hairpin)
            return child;            
        else if (child->type == interior || child->type == bulge)
        {
            Loop *prev_child = NULL;
            return getLeftmostChild (child, prev_child, level-1, metLevel);
        }    
        else
        {
            metLevel = 0;
            return child->Parent;
        }    
    }    
    return NULL;
}

/////////////////////
int Loop::saveDP (int &index, int begin, int end, char *tseq, char *tstr, char *tres, int resBegin, int resEnd)
{
    int j;
    for (j = begin; j <= end; j++)
    {
        char c = Input->CSequence[j];        
        if (c!='A' && c!='C' && c!= 'G' && c!='U' && c!='T' &&
            c!='a' && c!='c' && c!= 'g' && c!='u' && c!='t')
        {
            //printf ("Base: %c\n", c);    
            printf ("Unusual/unknown base - ignore loop\n");
            cout << "================================\n";            
            return 0;
        }    
        tseq[index] = c;        
        if (Input->BasePair(j) > j)
            tstr[index] = '(';
        else if (Input->BasePair(j) <= 0)
            tstr[index] = '.';
        else tstr[index] = ')';            
        if (resBegin && j == begin)
        {
            if (Input->BasePair(j) > j) tres[index] = '(';
            else if (Input->BasePair(j) < j) tres[index] = ')';            
            else tres[index] = '_';
        }    
        else tres[index] = '_';
        if (resEnd && j == end)
        {
            if (Input->BasePair(j) > j) tres[index] = '(';
            else if (Input->BasePair(j) < j) tres[index] = ')';            
            else tres[index] = '_';
        }            
        index++;
    }
    return 1;
}


////////////////////
void Loop::getExternalEnds (Loop *extLoop, Loop* endStem, int &begin, int &end)
{
    // find begin and end;
    Loop *tmp;
    tmp = extLoop->RightChild;
    if (tmp == endStem)
    {
        end = Input->Size;
        tmp = tmp->LeftSibling;
        if (tmp == NULL)
            begin = 1;
        else
            begin = tmp->end + 1;    
    }
    else
    {
        Loop *RightSibling = tmp;
        while (tmp->LeftSibling != endStem)
        {
            tmp = tmp->LeftSibling;
            RightSibling = tmp;
        }
        end = RightSibling->begin-1;
        tmp = tmp->LeftSibling;    // now tmp is endStem
        // now look for the 5' end (i.e. begin)
        tmp = tmp->LeftSibling;
        if (tmp == NULL)
            begin = 1;
        else
            begin = tmp->end + 1;                
    }
}


////////////////////
void Loop::extractHairpinFragments (int format, int level)
{
    int j, metLevel;
    Loop *loop, *parent, *child, *previousChild, *endStem;    
    previousChild = NULL;
    cout << "***" << Name << Num << " (" << begin << "," << end << ")***\n" ;
    // write bpseq format
    // first display the stem
    parent = getParent (this, endStem, level, metLevel);
    if (parent == NULL)
    {
        printf ("NULL parent - ignore hairpin\n");
        cout << "================================\n";
        return;    
    }   
    if (!metLevel)
    {
        printf ("Level not met - ignore hairpin loop\n");
        cout << "================================\n";
        return;    
    }
             
    if (format == BPSEQ)    
    {
        int shift1;
        shift1 = parent->begin-1;
        for (j = parent->begin; j <= parent->end; j++)
        {
            cout << j - shift1 << " ";
            cout << Input->CSequence[j] << " ";
            cout << MAX (0, Input->BasePair(j) - shift1);
            if (j == parent->begin || j == parent->end) cout << " *\n";
            else cout << "\n";
        }
    }        
    else if (format == DP)
    {
        char tseq[parent->end - parent->begin+2];    // temporary sequence array
        char tstr[parent->end - parent->begin+2];    // temporary structure array
        char tres[parent->end - parent->begin+2];    // temporary restriction array
        int index = 0;
        // parent is stackloop (level 1 or 2) or external (level 2)
        // not true any more -> for hairpin loops, it doesn't make much sense to continue with internal loops - 
        //                      they will be captured by the internal loop case
        if (parent->type == stackloop)    // level == 1 && 
        {
            if (!saveDP (index, parent->begin, parent->end, tseq, tstr, tres, 1, 1))
                return;            
        }
        else if (parent->type == external)
        {
            int begin, end;
            getExternalEnds (parent, endStem, begin, end);
            if (!saveDP (index, begin, end, tseq, tstr, tres, 0, 0))
                return;            
        }            
        else
        {
            printf ("Bad parent type - ignore hairpin\n");
            cout << "================================\n";
            return;                        
        }
        tseq[index] = '\0';
        tstr[index] = '\0';
        tres[index] = '\0';
        printf ("Seq: %s\nStr: %s\nRes: %s\n", tseq, tstr, tres); 
    }
    cout << "================================\n";
}

//////////////////////////////
void Loop::extractInternalBulgeFragments (int format, int level)
{
    int j, metLevelParent, metLevelChild;
    Loop *loop, *parent, *child, *previousChild, *endStem;    
    previousChild = NULL;
    cout << "***" << Name << Num << " (" << begin << "," << end << ")***\n" ;
    parent = getParent (this, endStem, level, metLevelParent);
    if (parent == NULL)
    {
        printf ("NULL parent - ignore internal loop or bulge\n");
        cout << "================================\n";
        return;    
    }     
    child = getLeftmostChild (this, previousChild, level, metLevelChild); 
    if (child == NULL)    //child->type != stackloop && child->type != hairpin)
    {
        printf ("NULL child - ignore internal loop or bulge\n");
        cout << "================================\n";
        return;
    }
    if (!metLevelParent && !metLevelChild)
    {
        printf ("None of the levels met - ignore internal loop or bulge\n");
        cout << "================================\n";
        return;    
    }
    if (format == BPSEQ)
    {
        int shift1, shift2;                
        shift1 = parent->begin-1;
        shift2 = parent->begin + child->end - child->begin - 4;            
        for (j = parent->begin; j <= child->begin + 1; j++)
        {
            printf ("%d %c %d", j - shift1, Input->CSequence[j], 
                MAX (0, Input->BasePair(j) - shift2));
            if (j == parent->begin || j == child->begin+1) printf (" *\n");
            else printf ("\n");
        }
        for (j = child->end-1; j <= parent->end; j++)
        {
            printf ("%d %c %d", j-shift2, Input->CSequence[j],
                MAX (0, Input->BasePair(j) - shift1));
            if (j == child->end-1 || j == parent->end) printf (" *\n");
            else printf ("\n");    
        }
    }
    else if (format == DP)
    {
        char tseq[parent->end - parent->begin+2];    // temporary sequence array
        char tstr[parent->end - parent->begin+2];    // temporary structure array
        char tres[parent->end - parent->begin+2];    // temporary restriction array
        int index = 0;
        // parent type is stackloop (level 1 or 2) or external loop (level 2)
        // child type is stackloop (level 1 or 2) or hairpin loop (level 2)
        if (parent->type == stackloop && child->type == stackloop)
        {
            if (!saveDP (index, parent->begin, child->begin + 1, tseq, tstr, tres, 1, 1))
                return;            
            if (!saveDP (index, child->end-1, parent->end, tseq, tstr, tres, 1, 1))
                return;            
        }
        if (parent->type == stackloop && child->type == hairpin)
        {
            if (!saveDP (index, parent->begin, child->begin, tseq, tstr, tres, 1, 0))
                return;            
            if (!saveDP (index, child->begin+1, parent->end, tseq, tstr, tres, 0, 1))
                return;            
        }        
        else if (parent->type == external && child->type == stackloop)
        {
            // parent must be external loop
            int begin, end;
            getExternalEnds (parent, endStem, begin, end);
            if (!saveDP (index, begin, child->begin+1, tseq, tstr, tres, 0, 1))
                return;            
            if (!saveDP (index, child->end-1, end, tseq, tstr, tres, 1, 0))
                return;                    
        }                    
        else if (parent->type == external && child->type == hairpin)
        {
            // parent must be external loop
            int begin, end;
            getExternalEnds (parent, endStem, begin, end);
            if (!saveDP (index, begin, child->begin, tseq, tstr, tres, 0, 0))
                return;            
            if (!saveDP (index, child->begin+1, end, tseq, tstr, tres, 0, 0))
                return;            
        }        
        tseq[index] = '\0';
        tstr[index] = '\0';
        tres[index] = '\0';
        printf ("Seq: %s\nStr: %s\nRes: %s\n", tseq, tstr, tres);         
    }
    cout << "================================\n";
}

//////////////
void Loop::extractMultiFragments (int format, int level)
{
    int j, metLevel, totalMetLevel;
    Loop *loop, *parent, *child, *previousChild, *endStem;    
    previousChild = NULL;
    totalMetLevel = 0;
    int parentEnd, restrictStart, restrictEnd;
    cout << "***" << Name << Num << " (" << begin << "," << end << ")***\n" ;
    parent = getParent (this, endStem, level, metLevel); 
    totalMetLevel += metLevel;      
    if (parent == NULL)
    {
        printf ("NULL parent - ignore multi loop\n");
        cout << "================================\n";
        return;    
    }
        
    child = getLeftmostChild (this, previousChild, level, metLevel);
    if (child == NULL)
    {
        printf ("NULL child - ignore multi loop\n");
        cout << "================================\n";
        return;
    }    
    
    totalMetLevel += metLevel;
    if (format == BPSEQ)
    {
        printf ("BPSEQ format not implemented yet\n");        
    }    
    else if (format == DP)
    {
        int start, end;
        char tseq[parent->end - parent->begin+2];    // temporary sequence array
        char tstr[parent->end - parent->begin+2];    // temporary structure array
        char tres[parent->end - parent->begin+2];    // temporary restriction array
        int index;
        index = 0;
        if (parent->type == external)
        {
            getExternalEnds (parent, endStem, start, parentEnd);
            restrictStart = 0;
        }    
        else
        {
            start = parent->begin;
            restrictStart = 1;
        }            
        while (child != NULL)
        {
            if (level == 1 && child->type != stackloop)
            {
                printf ("Non-stem child - ignore multi loop\n");
                cout << "================================\n";
                return;
            }
            else if (level == 2 && child->type != stackloop && child->type != hairpin)
            {
                printf ("Non-stem child - ignore multi loop\n");
                cout << "================================\n";
                return;            
            }
            if (child->type == hairpin)
            {
                restrictEnd = 0;
                end = child->begin;
            }
            else
            {
                restrictEnd = 1;
                end = child->begin + 1;
            }    
            if (!saveDP (index, start, end, tseq, tstr, tres, restrictStart, restrictEnd))
                return;            
            if (child->type == hairpin)
            {
                restrictStart = 0;
                start = child->begin+1;
            }
            else
            {
                restrictStart = 1;
                start = child->end-1;
            }                                
            child = getLeftmostChild (this, previousChild, level, metLevel);              
            totalMetLevel += metLevel;
        }     
        if (parent->type == external)
        {
            end = parentEnd;
            restrictEnd = 0;
        }    
        else
        {
            end = parent->end;
            restrictEnd = 1;    
        }                    
        if (!saveDP (index, start, end, tseq, tstr, tres, restrictStart, restrictEnd))
            return;                        
        tseq[index] = '\0';
        tstr[index] = '\0';
        tres[index] = '\0';
        if (!totalMetLevel)
        {
            printf ("None of the levels met - ignore multi loop\n");
            cout << "================================\n";
            return;    
        }          
        printf ("Seq: %s\nStr: %s\nRes: %s\n", tseq, tstr, tres); 
    }
    cout << "================================\n";
}

//////////////
void Loop::extractFragments (LoopType inputType, int level, int format)
// 
{
    int j, shift1, shift2;
    Loop *loop, *parent, *child, *previousChild;    
    previousChild = NULL;
    if (type == inputType)
    {
        if (type == hairpin)
        {    
            extractHairpinFragments (format, level);
        }        
        else if (type == bulge || type == interior)
        {
            extractInternalBulgeFragments (format, level);
        }
        else if (type == multi)
        {
            extractMultiFragments (format, level);
        }        
    }
    loop = RightChild;
    if (loop)
    {
        do 
        {
            loop->extractFragments (inputType, level, format);
            loop = loop->LeftSibling;
        } while (loop != NULL);
    };    
}



//////////////////////
void Loop::extractFragmentsLevel3 (Loop *startLoop, int format)
// startLoop is the first loop of a new fragment
// split at multi-loops, external loops and pseudoknots
// assume DP format
{
    int begin1, end1, begin2, end2;
    Loop *branch, *node;
    if (startLoop->type == external || startLoop->type == multi || startLoop->type == pseudo)
    {
        branch = startLoop->RightChild;
        while (branch != NULL)
        {
            extractFragmentsLevel3 (branch, format);
            branch = branch->LeftSibling;
        }    
    }
    else
    {    
        begin1 = startLoop->begin;
        end1 = startLoop->end;        
        char tseq[end1 - begin1+2];    // temporary sequence array
        char tstr[end1 - begin1+2];    // temporary structure array
        char tres[end1 - begin1+2];    // temporary restriction array
        int index = 0;        
        node = startLoop->RightChild;
        while (node != NULL)
        {
            if (node->type == multi || node->type == external || node->type == pseudo)
            {
                // stop
                cout << "***" << node->Name << node->Num << " (" << node->begin << "," << node->end << ")***\n" ;
                begin2 = node->begin;
                end2 = node->end;
                if (node->type == pseudo)
                {
                    // the end of a pseudoknot don't make a base pair
                    begin2 = node->Parent->begin;
                    end2 = node->Parent->end;                    
                }
                if (!saveDP (index, begin1, begin2, tseq, tstr, tres, 1, 1))
                    return;            
                if (!saveDP (index, end2, end1, tseq, tstr, tres, 1, 1))
                    return;   
                tseq[index] = '\0';
                tstr[index] = '\0';
                tres[index] = '\0';
                printf ("Seq: %s\nStr: %s\nRes: %s\n", tseq, tstr, tres);         
                cout << "================================\n";                          
                extractFragmentsLevel3 (node, format);   
                break;                          
            }
            else if (node->type == hairpin)
            {
                cout << "***" << node->Name << node->Num << " (" << node->begin << "," << node->end << ")***\n" ;
                if (!saveDP (index, begin1, end1, tseq, tstr, tres, 1, 1))
                    return;                        
                tseq[index] = '\0';
                tstr[index] = '\0';
                tres[index] = '\0';
                printf ("Seq: %s\nStr: %s\nRes: %s\n", tseq, tstr, tres);         
                cout << "================================\n";       
                break;                               
            }
            else
            {
                node = node->RightChild;            
            }    
        }    
    }        
}


//////////
void Loop::level3Recursion (Loop *loop, int *begin, int *end, int &num_cuts, int format)
{
    if (loop == NULL)
        return;
    if (loop->type == pseudo)
    {
        begin[num_cuts] = loop->Parent->begin;
        end[num_cuts] = loop->Parent->end;
        num_cuts++; 
        extractFragmentsLevel4 (loop, format);
    }
    else
    {
        level3Recursion  (loop->RightChild, begin, end, num_cuts, format);
        level3Recursion  (loop->LeftSibling, begin, end, num_cuts, format);        
    }
}


//////////////////////
void Loop::extractFragmentsLevel4 (Loop *startLoop, int format)
// startLoop is the first loop of a new fragment
// split at pseudoknots only
// assume DP format

// not functional yet
{
    
    Loop *branch, *node, *previousChild;
    previousChild = NULL;
    if (startLoop->type == pseudo)
    {
        branch = startLoop->RightChild;
        while (branch != NULL)
        {
            extractFragmentsLevel4 (branch, format);
            branch = branch->LeftSibling;
        }    
    }
    else
    {    
        int num_cuts = 0;
        int begin[100];
        int end[100];
        char tseq[startLoop->end - startLoop->begin + 2];    // temporary sequence array
        char tstr[startLoop->end - startLoop->begin + 2];    // temporary structure array
        char tres[startLoop->end - startLoop->begin + 2];    // temporary restriction array
        int index = 0;        
        int i;          
        int restrictedHead;
        if (startLoop->type == external)
            restrictedHead = 0;
        else restrictedHead = 1;

        if (startLoop->begin == 849)
        {
            printf ("Stop\n");
        }        
        
        // do recursion on the right child. The left sibling is another story, we don't have to go there now.
        level3Recursion  (startLoop->RightChild, begin, end, num_cuts, format);        

        cout << "***pseudo (" << startLoop->begin << "," << startLoop->end << ")***\n" ;       

        if (num_cuts == 0)     // there's no pseudoknot
        {
            if (!saveDP (index, startLoop->begin, startLoop->end, tseq, tstr, tres, restrictedHead, restrictedHead))
                return;
            tseq[index] = '\0';
            tstr[index] = '\0';
            tres[index] = '\0';
            printf ("Seq: %s\nStr: %s\nRes: %s\n", tseq, tstr, tres);
            cout << "================================\n";
        }
        else     // there's at least one cut
        {
            if (!saveDP (index, startLoop->begin, begin[num_cuts-1], tseq, tstr, tres, restrictedHead, 1))
                return;

            for (i=num_cuts-2; i >= 1; i--)
            {
                if (!saveDP (index, end[i], begin[i-1], tseq, tstr, tres, 1, 1))
                    return;
            }
            if (!saveDP (index, end[0], startLoop->end, tseq, tstr, tres, 1, restrictedHead))
                return;

            tseq[index] = '\0';
            tstr[index] = '\0';
            tres[index] = '\0';
            printf ("Seq: %s\nStr: %s\nRes: %s\n", tseq, tstr, tres);
            cout << "================================\n";
        }
    }        
}


