//#include "Sequence.h"
#ifndef LOOP_H
#define LOOP_H

#include "Input.h"
#include "Bands.h"
#include "LoopList.h"

#include <stack>
#include <math.h>

class Loop 
{

 public:

  //variables:

  T_IntList *ILoops, *MLoops;
  Stack * St;

  Loop* RightChild, *LeftSibling, *Parent, *PseudoParent; //Rigth Child, Left Sibling, Parent and the nearest PseudoLoop Parent of the node
  int begin, end;				          //The closed region [begin,end] assosicated with the root
  LoopType type;
  ReadInput * Input;
  Bands  * bandpattern;				          // For bands


  int NumberOfChildren;
  int NumberOfUnpaird;					  //Number if unpaired bases within the loop
  int Pseudo_NumBranches;				  //Number of Branches by considering pseudoknot children as 2.

  PseudoNestedType nested;                //Where it is nested in a pseudoknot
  int NumberOfUnBandChild;
  int NumberOfBandChild;
  int CurrentBandRegion;                 //Current band region that we are checking, starting from end
  int NumberOfBands;

  vector <int> bandborders;  //the borders of bands involved in this paticular pseudo region
  vector <int> bands_remove; //bands to be removed from this particular pseudo region in order to render it pseudoknot free

  string Name;        //added by Vera (5 different variables)
  int Num;
  int lengths[MaxNlength];
  int size_lengths;
  
  int tot_fb;
  int fb[MaxBases];
  
  int tot_cbp;
  int cbp[MaxPairs];

  //functions:

  Loop(int b, int e, ReadInput * R, Bands * B, Stack * S); //Constructor
  ~Loop();

  void addLoop(int begin, int end);		//Adds a loop [begin,end] to the tree
  void Print(int i);		   //Prints the loop tree in a hierarchical structure
  //  void setTypes();		   //Set the type of each Loop in the tree(interior, hairin, pseudo, ...)
  void loopSetType();		   // set the type of [begin,end]
  void setPseudoParent();
  //  void FindBands();		   //Finds the bands if the loop is pseudoknotted (and for all of its children as well)
  void LoopFindBands();		   //Finds the bands if the loop is pseudoknotted

  int NumUnpaired();			//Number of Unpaired bases in this loop
  void FindInnerLoops(int border1, int border2);		//Finds all Interior nodes in the band

  //Added by Baharak
  void PseudoNestedCheck();			    //Check where in a pseudoknot its children located
  void WhereLocated();						//Find the locatin of the loop in a pseudoknot
  void PropagateNested(); //propagate the location to the children

  // function added by Mirela - started on April 14, 2005
  void extractHairpinFragments (int format, int level);
  void extractInternalBulgeFragments (int format, int level);
  void extractMultiFragments (int format, int level);
  void extractFragments (LoopType inputType, int level, int format);
  int saveDP (int &index, int begin, int end, char *tseq, char *tstr, char *tres, int resBegin, int resEnd);
  void getExternalEnds (Loop *extLoop, Loop* endStem, int &begin, int &end);
  void extractFragmentsLevel3 (Loop* startLoop, int format);
  void extractFragmentsLevel4 (Loop* startLoop, int format);
  void level3Recursion (Loop *loop, int *begin, int *end, int &num_cuts, int format);

  //Added by Vera
  //note connection functions handle coordinates, length(s), and connections
  
  Loop * StemConnections(int index);
  
  void StemContext(int pos1, int cur_pair); //finds the context of non canonical base pair
  
  void SingleStacked(int index); //deals with a stem of length 1

  void HairpinConnections(int index);
  void HairpinType(); //this is according to a motif.txt

  void MultiConnections(bool external, int index);
  void MultiAsymmetry();
  void TypeFreeBase(int i, bool external);
  void IncFreeBase();
  void ClosingBasePair(int k, bool external);
  void IncCBP();

  void Numbers(int pos1,int pos3,bool nonband); //calculate number of bases, and base pairs involved in a stem (plus their corresponding %s)
  int BPRec(int pos1, int& c, int& d);
  
  void PseudoType(); //to find out what type of pseudoknot we are looking at, 
  //plus  other things mentioned in the comments of the actual impelementation
  bool NotInChild(int pos, int* children, int child_size);

  //Mirela's funcs
  int standard_base(char nucleotide);
  //this is to differentiate between standard and non standard nucleotides
  
  int nuc_to_int(char nucleotide);
  // PRE:  nucleotide is 'A', 'C', 'G' or 'U'
  // POST: Return 0 for A, 1 for C, 2 for G, 3 for U
  
  //this one slighly modified by Vera
  double StemFreeEnergy(int pos1, int pos2);
  // PRE:  Seq1 and seq2 are 2 (almost) complementary sequences forming a stem,
  //       and read from 5' to 3'. Seq1 and seq2 must have the same length.
  // POST: Returns the free energy of the stem, according to Turner stacking parameters.
  //       Non-canonical pairs are considered 0.  

};

#endif
