#include "Stack.h"
#include "Loop.h"
#include "Bands.h"
#include "Input.h"
#include "PlotRna.h"

#include "oentry.h"
#include "clentry.h"
#include "interval.h"


#include <list>


vector <int> pos;
vector <char> bases;
bool alternatives=false;
vector <int> to_be_removed;
int rem_set=0;
ReadInput * R; 
Loop *L;
int split = 1;
int free_bases = 0;

//#############################################################################

//helper functions for printing different structures

template<class type>
void print(const vector<type>& v)
{
  for (unsigned int counter = 0; counter < v.size(); counter++) {
    cout << v.at(counter) << " ";
  }
  cout << endl;
}

void printarray (int arg[], int length)
{
  for (int n=0; n<length; n++)
    cout << arg[n] << " ";
  cout << endl;
}

void printlist(const list<int>& lis)
{
  list<int>::const_iterator LI;
  for(LI=lis.begin(); LI != lis.end(); LI++)
    {
      cout << (*LI) << " ";
    }
  cout << endl;
}

//******************************************************************************************
//recursive function for figuring out the max independent set of arcs in out molecule

vector<int>  circle_graphs(int first, int last, int size, bool first_call)
{
  int free_b=((last-first+1)-size*2);
  int intervals=size;
  //NOTE: from hereon after, arc=interval :)

  // now let's do the DP for calculating the min num of base pairs that needs to be broken
  int num_pairs=1;//have to start from 1 not zero, otherwise there will be confused with unpaired bases later on
  int alpha[last-first+1];

  for(int y=0; y<(last-first+1); y++)
    {
      if((pos[first+y] > (first+y))&&((first<=pos[first+y]-1)&&(pos[first+y]-1<=last))) //open brakets bases only that are within the interval
    {
      alpha[pos[first+y]-(first+1)]=num_pairs;
      alpha[y]=num_pairs;
      num_pairs++;  
    }
      else if ((pos[first+y]==0) ||((first>pos[first+y]-1)||(pos[first+y]-1>last))) //ignored or free bases
    alpha[y]=0;
    }

/*
    cout << "ALF:  ";
    printarray(alpha, last-first+1);
*/

  //let's calculate CMIS now ;)
  //CMIS[i]=weight of interval i + all nested intervals
  //MIS[i]=CMIS[i] + weight of all the intervals preceeding i that have been closed
  //overall, the algorithm should take roughtly O(dnlogn), where d is the density 
  //(ie max number of nested arcs in the molecule), n is the number of intervals
  //Note: implement Open, as a list as suggested in the paper
  // clist is a list as well
  list<oentry> Open;

  int i, l, r, ww, tempi, templ;
  bool encountered = false;
  i=r=l=-1;
  int CMIS[intervals];
  
   
  for(int m=first; m<=last;m++)
    {
      if((pos[m] !=0)&&((first <= pos[m]-1)&&(pos[m]-1 <=last)))
    {
      i = alpha[m-first]-1;
      //if dealing with an open brace, then push i onto Open, set clist[i] to null, and CMIS[i]=0
      if(pos[m] > m)
        {
          //get correspoinding interval
          oentry oe = oentry(i);
          Open.push_back(oe);
          CMIS[i]=0;
        }
      else // no free bases, thus this is a closing brace
        {
          l = pos[m]-(first+1);
          r = m-first;
          //printf ("position %d should be a closing brace\n");

          //add the weight of the interval 
          CMIS[i] += 1; //as all intervals have a weight of 1 
          //pop the oentry for this interval from Open AND find x's with corresponding ww's
          //for all x's in Open, s.t. their i < then i of our m,
          //ie they have been opened before our currently looked at interval
          //was opened, find r in clist(x), where (r,ww) is an entry and
          //r=max{r<l}, else ww=0 (if r DNE)

          bool Popped = false;

          list<oentry>::iterator theIterator;
          for(theIterator = Open.begin(); (!Popped)&&(theIterator != Open.end()); theIterator++)
        {
          if(i == (*theIterator).getI())
            {
              Open.erase(theIterator); // pop oentry
              Popped = true;
            }
        }

          list<oentry>::iterator theIt;
          for(theIt = Open.begin(); theIt != Open.end(); theIt++)
        {
          tempi = (*theIt).getI();
          for(int y=0; (!encountered) && y<(last-first+1); y++) // search for the left end point, it should be the first encountered
            {
              if((alpha[y]-1)==tempi) // the first brace we encounter will be the open one, ie the left end point :)
            {
              encountered = true;
              templ = y;
              //cout << "the tempi and templ " << tempi << " " << templ << endl;
            } 
            }
          if ((encountered) && (templ<l))// then we have found our x :)
            {
              ww=0;
              int best_r, tr;
              best_r = tr = -1;

              list<clentry>::iterator Iterator2;
              for(Iterator2=(*theIt).getCLIST_IT(); Iterator2 != (*theIt).at_end(); Iterator2++)
            {
              tr = (*Iterator2).getREP();
              if((tr > best_r)&&(tr < l))
                {
                  ww = (*Iterator2).getWW();
                  best_r = tr;
                }
            }
              //correct CMIS for each x (ie our closed interval had its impact)
              CMIS[tempi] = max(CMIS[tempi],CMIS[i]+ww);
              //push onto clist of each of our x's (right hand point of i, CMIS[x])        
              clentry cl = clentry(r, CMIS[tempi]);
              (*theIt).push_cl(cl);
              ww=0;
            }
          encountered = false;
        }
        }
    }
    }

/*  
    cout << "CMIS: ";
    printarray(CMIS, intervals);
*/  

  //let's calculate MIS now
  int MIS[(intervals*2)+free_b];
  int tempMIS = 0;

  for(int m=first; m<=last; m++)
    {
      if(((pos[m] > m)||(pos[m]==0))||(((pos[m]-1)<first)||((pos[m]-1)>last)))
    {
      if(m-(first+1) >= 0)
        {
          MIS[m-first] = MIS[m-first-1];
        }
      else
        {
          MIS[m-first]=0;
        }
    }
      else // thus we are looking at a closing braket, MIS[m]=max{MIS[m-1], MIS[(left point of i)-1]+CMIS[i]}
    {
      //since close brace can't be located at 0, we can foget about error precausions, ie can take MIS[m-1]
      i = alpha[m-first]-1;
      l = pos[m]-(first+1);
      tempMIS = 0;

      if(l != 0)
        {
          tempMIS = MIS[l-1];
        }
      MIS[m-first] = max(MIS[m-(first+1)], (tempMIS+CMIS[i]));
    }
    }

/*  
    cout << "MIS:  ";
    printarray(MIS, ((intervals*2)+free_b));
*/  

  //(total # of arcs to be removed)=(total # of arcs)-MIS[num_paired-1];
  int to_remove = (intervals)-MIS[((intervals*2)+free_b)-1];
  //cout <<"Total number of arcs to be removed: " << to_remove << endl;

  int UNREMOVED=1;
  vector<int> remove;
  //BACKTRACKING, LET THE FUN BEGIN
  if(to_remove !=0) //if nothing needs to be removed, then we should not need to do backtracking 
    {
      //now let's backtrack and finally figure out what needs to be removed
      // pk-pseudoknot(s) 
      int weight=0;
      for(int h=0; h<intervals; h++)
    {
      remove.push_back(1);
    }
    
      UNREMOVED=0;
  
      for(int m=last; m >= first; m--)
    {
      //only worry about the close brakets
      if(((pos[m] !=0)&&(first<=pos[m]-1))&&(pos[m]-1<=last))
        {
          if(pos[m]-1 < m)
        {
          i = alpha[m-first]-1;
          l = pos[m]-1;
          r = m;
          tempMIS = 0;
    
          if((pos[m]-(first+1)) != 0) // this is so we won't get outside of the array bounds and try to access something like MIS[-1]
            {
              tempMIS = MIS[(pos[m]-(first+1))-1];
            }
          if(MIS[m-(first+1)] <= (tempMIS+CMIS[i])) //ie second choice or a tie
            {      
              if(MIS[m-(first+1)]==(tempMIS+CMIS[i]))
            alternatives=true;                      
          
              remove[i]=0;
              UNREMOVED=1;
              int inside_size=0;
              //need to figure out how many intervals are inside the currently being looked at
              //interval, and if there intervals inside, then make sure we know exactly which ones
              //so that when we get our remove vector back from the circle_graph(), we know which
              //intervals in our alpha need to be unremoved.
              list<int> inside_int;

              for(int b=l+1; b<r; b++)
            {
              if(((pos[b]-1>b)&&(l<pos[b]-1))&&(pos[b]-1<r))
                {
                  inside_size++;
                  inside_int.push_back(alpha[b-first]-1);
                }
            }
          
              if(inside_size)
            {
              //inside_int.sort();
              vector <int> inside;
              inside = circle_graphs(l+1,r-1, inside_size, false);
              if(inside[inside_size])
                {
                  for(int k=0; k<inside_size; k++)
                {
                  if(!inside[k])
                    {
                      remove[inside_int.front()]=0; 
                    }
                  inside_int.pop_front();
                }
                }
            }
              m=pos[m]-1;
            }
          // else, we made the first choice, then just go to m-1, the loop will do that for us
        }
        }
    }
    }   
  else
    {
      for(int h=0; h<intervals; h++)
    {
      remove.push_back(0);
    }
    }
  remove.push_back(UNREMOVED);
  /*
    cout<< "OUR CURRENT REMOVE:\n";
    print(remove);
  */
  
  if(first_call)
    {
      for(int i=0; i<remove.size()-1; i++)
    {
      if(remove[i])
        {
          rem_set++;
          for(int j=0; j<=last-first; j++)
        {
          if(alpha[j]==i+1)
            {
              to_be_removed.push_back(j+1);
              break;
            }
        }
        }
    }
    }
  
  return remove;
}

void readStructure_ct (char *filename)
// must fill variables:
//  - free_bases
//  - bases (vector)
//  - pos (vector)
{
    char junk[2560];
    ifstream fileStr;
    int position;
    char base;
    int posm1;
    int posp1;
    int pair;
    int pos2;
    fileStr.open(filename);
    if (!fileStr.is_open())
    { cout << "Error opening the input file " << filename << endl; exit (1); }
    while(fileStr.peek() == '#')
    {
        fileStr.getline(junk, MAXLINE);
    }

    // changed to ct format. We have one more line (the energy line) that we have to ignore.
    fileStr.getline(junk, MAXLINE);

    while (!fileStr.eof())
    {
        fileStr.getline(junk, MAXLINE);
        //cout << junk << endl;
        if (sscanf (junk, "%d %c%d%d%d%d", &position, &base, &posm1, &posp1, &pair, &pos2) == EOF)
            break;
        bases.push_back(base);
        pos.push_back(pair);
        if (pair == 0)
            free_bases++;
        //printf ("%d, %c, %d, %d, %d, %d\n\n", position, base, posm1, posp1, pair, pos2);
    }
    
    fileStr.close();    
}

//################################################################################################


void readStructure_bpseq (char *filename)
{
  char junk[256];
  char fileSeq[100], fileStruct[100], outPSFile[100];
  char prefix[100];

  strcpy(prefix, filename);
  strcpy(fileStruct, prefix);
  strcpy(fileSeq, prefix);
  strcpy(outPSFile, prefix);
  //strcat(fileStruct, ".bpseq");   
  strcat(outPSFile, ".ps");
  
  //cout << "Our files: " << fileStruct << " " << outPSFile << endl;

    //if this gives you segmentation fault, then comment this block out up to "the end" comment 
  //and uncomment the second version of the file reading for loop
  ifstream fileStr;
  fileStr.open(fileStruct);
  if (!fileStr.is_open())
  { cout << "Error opening the input file"; exit (1); }
  
  //get the comments marked by '#' at the beginning of the line out
  //remove comments from file
  while(fileStr.peek() == '#')
    {
      fileStr.getline(junk, MAXLINE);
    }

    // changed to ct format. We have one more line (the energy line) that we have to ignore.
    //fileStr.getline(junk, MAXLINE);

  char rna_m;
  bool poz = true;   
  bool base = false;
  bool pushed = false;
  bool breaked = false;
  int num_bases = 0;
  int rez = 0;
  int num_pos = 0;
  free_bases=0;

  int starttime,currtime,timeused; 
  starttime = clock();
  
  // fill the vectors pos and bases
  while((!fileStr.eof()) && fileStr.good())      
    {
      rna_m = fileStr.get();

      currtime = clock();
      timeused = (currtime - starttime) / 1000;
      if(timeused > 25000)
    {
      cout << "Something's wrong with the input file (not in correct bpseq format?). I can't parse it. Exiting the program.\n"; 
      exit (1);
    }

      if(fileStr.eof())
    break;
      if(((rna_m != ' ')&&(rna_m != ','))&&(((rna_m != '\t')&&(rna_m != '\n'))&&(rna_m != '\r')))
    {
      if(poz && (!pushed))
        {        
          base = true;
          poz = false;
          pushed = true;
          while(((fileStr.peek() != ' ')&&(fileStr.peek() != ','))&&(fileStr.peek() != '\t'))
        {
          rna_m = fileStr.get();
          
          currtime = clock();
          timeused = (currtime - starttime) / 1000;
          if(timeused > 25000)
            {
              cout << "Something's wrong with the input file (not in correct bpseq format?). I can't parse it. Exiting the program.\n"; 
              exit(1);  
            }
        }
        }
      if(base && (!pushed))
        {
        if((rna_m=='t')||(rna_m=='u')||(rna_m=='T'))
        rna_m='U';
        if((rna_m<='z')&&(rna_m>='a'))
        rna_m = toupper(rna_m);

          //cout << "Push base: " << rna_m << endl;  
          bases.push_back(rna_m);
          num_bases++;
          base = false;
          pushed = true;
        }
      if(!pushed)
        {
          poz = true;
          {
        if((fileStr.peek() == '\n')||(fileStr.peek() == '\r'))
          {            
            pos.push_back(rna_m-48);
            if(rna_m-48==0)
              free_bases++;
          }
        else 
          {
            rez = 0;
            while ((fileStr.peek() != ' ') && (fileStr.peek() != '\n'))
              {
            currtime = clock();
            timeused = (currtime - starttime) / 1000;
            //cout << "Time used: " << timeused << " and the boolean is " << (timeused>25000) << endl;
            if(timeused > 25000)
              {
                cout << "Something's wrong with the input file (not in correct bpseq format?). I can't parse it. Exiting the program.\n"; 
                exit (1);
              }
            rez = (rna_m-48) + rez*10;
            rna_m = fileStr.get();
            if(fileStr.eof())
              {
                break;
                breaked = true;
              }
              }
            if(!breaked)
              rez = (rna_m-48) + rez*10;              
            pos.push_back(rez);
          }
          }
        }
      pushed = false;
    }
    }  
  fileStr.close(); 
}

// ###########################################################


void findNumArcsToRemove ()
{

  //cout << "Free bases: " << free_bases << endl
  cout << "Strand Length: " << pos.size() << endl
       << "##########################################################################################\n\n";


  if(pos.size()==0)
    {
      cout << "For some reason the strand length is zero. This is not allowed.\n";
      cout << "Please, check the file name that you have provided me with (I don't believe it exists) and try again.\n";
      exit(1);
    }

  //now figure out if there are pseudoknots, and the min number of arcs to be broken to make it
  //pseudoknot free 
  
  int rem_size=(pos.size()-free_bases)/2;
  vector<int>rem=circle_graphs(0, (pos.size())-1, rem_size, true);
   
  cout << "Total number of arcs to be removed: " << rem_set << endl;
  
  if(rem_set)
  {
  cout << "Specific arcs to be removed:\n";
  for(int g=0; g< rem_size; g++)
    {
      if(rem[g])
    {
      cout << g << " ";
    }
    }
    cout << endl;
}  
  
  if(alternatives)
    cout << "However, there are alternative ways to make the structure pseudoknot free\n";
    
 
  /*
    cout << "THE ARRAY TO BE REMOVED: ";
    for(int i=0; i<to_be_removed.size(); i++)
    cout << to_be_removed[i] << ",";
    cout << endl;
  */
}  




////////////////////////////////////////////////////////////////////////////////////

void buildTree ()  
  //proceed with the rest of the analyser's activities:
  //first do some preliminary initalizations
{    

  int z = bases.size();
  R = new ReadInput(z, pos, bases, to_be_removed);
  Stack * s = new Stack(R);
  Bands * B = new Bands(R, s);
  // Mirela: June 19, 2005: set the begin and end of the exterior loop as 1 and z, as sopposed to 0 and MaxN+1
  L = new Loop(1, z, R, B, s);
  
  //L = new Loop(0, MaxN+1, R, B, s);    //before this all the preliminary initialization is done,
  //after this we do the actual calculations for the stack, bands, and loop tree

  int a, b;

  for (int i = 1; i <= R->Size; i++) // walking through the whole molecule
    {
      if (R->BasePair(i)>= 0) //if we have a paired base
    {
      if (s->Add(i, a, b)) //if we figured that we had a closed region starting with a (either pseudoknotted or not).
        //This is where we mark stuff with B, P, N...
        {
          L->addLoop(a,b); //since its a closed region, add it to our tree. THIS IS WHERE MOST CALCULATION IS DONE, IMPORTANT!
        };
    };
    };
}

//#####################################################

void findNumBandsToRemove ()
{  
  int g=R->bands_to_remove.size();  
  cout << "Total number of bands to be removed: " << g/4 << endl;
  if(g/4)
    {
      cout << "Coordinates of the bands to be removed are:\n";
      for(int p=0; p<g-3; p++)
    {
      cout << R->bands_to_remove[p] << "-" << R->bands_to_remove[p+3] << ", " << R->bands_to_remove[p+1] << "-" << R->bands_to_remove[p+2];
      if(p != g-4)
        cout << "; ";
      p+=3;
    }
      cout << endl;
    }
}    



//#####################################################

int main(int argc, char ** argv)
{
      
    if (argc != 3 && argc != 4)
    {
        printf ("Usage: %s [-analyse|-split [1|2|3|4]] <file.ct or file.bpseq>\n", argv[0]);
        return 0;
    }     
    if (strcmp (argv[1], "-analyse") == 0)
    {
        if (strstr (argv[2], ".ct") != NULL)
            readStructure_ct (argv[2]);
        else if (strstr (argv[2], ".bpseq") != NULL)
            readStructure_bpseq (argv[2]);
        else
        {
            cout << "Filename has to be <file>.ct or <file>.bpseq\n";
            exit(1);
        }
        findNumArcsToRemove ();
        buildTree ();  
        findNumBandsToRemove();
        printf("##########################################################################################\n\n");
        printf ("-------------------------------THE LOOP TREE-------------------------------\n");
        L->Print(-1); 
        //print our tree, this is where all the actual calculation for our analyser are done/initialized        
    }
    else if (strcmp (argv[1], "-split") == 0)
    {
        int level = atoi (argv[2]);
        readStructure_ct (argv[3]); 
        buildTree ();      
        if (level <=2) 
        {
            L->extractFragments (hairpin, level, DP);
            L->extractFragments (bulge, level, DP);
            L->extractFragments (interior, level, DP);
            L->extractFragments (multi, level, DP);
        }
        else if (level == 3)
        {
            L->extractFragmentsLevel3 (L, DP);
        }
        else if (level == 4)
        {
            L->extractFragmentsLevel4 (L, DP);
        }
    }        


  //now plot our molecule via .ps file
/*  
  //this is to find out the molecule id (extract it from the file name of the input file)
   char * molecule_id = strrchr(argv[1], (int) '/');
   char delims[] = "/";
   molecule_id = strtok(molecule_id, delims);
   if(molecule_id==0)
     molecule_id = argv[1];
  
   PlotRna(bases, pos, outPSFile, molecule_id, to_be_removed);
*/
}

